import typing


class StratumRetentionPredicatePerfectResolution:
    """Functor to implement the perfect resolution stratum retention policy, for
    use with HereditaryStratigraphicColumn.

    This functor enacts the perfect resolution policy by specifying
    whether a stratum with deposition rank r should be retained within the
    hereditary stratigraphic column after n strata have been deposited.

    The perfect resolution policy retains all strata. So, comparisons between
    two columns under this policy will detect MRCA rank with zero
    uncertainty. So, MRCA rank estimate uncertainty scales as O(1) with respect
    to the greater number of strata deposited on either column.

    Under the perfect resolution policy, the number of strata retained (i.e.,
    space complexity) scales as O(n) with respect to the number of strata
    deposited.

    See Also
    --------
    StratumRetentionCondemnerPerfectResolution:
        For a potentially more computationally efficient specificiation of the
        perfect resolution policy that directly generates the ranks of strata
        that should be purged during the nth stratum deposition.
    """

    def __call__(
        self: 'StratumRetentionPredicatePerfectResolution',
        stratum_rank: typing.Optional[int]=None,
        num_stratum_depositions_completed: typing.Optional[int]=None,
    ) -> bool:
        """Decide if a stratum within the stratagraphic column should be
        retained or purged.

        Every time a new stratum is deposited, this method is called on each
        stratum present in a HereditaryStratigraphicColumn to determine whether
        it should be retained. Strata that return False are immediately purged
        from the column, meaning that for a stratum to persist it must earn a
        True result from this method each and every time a new stratum is
        deposited.

        This functor's retention policy implementation guarantees that columns
        will retain all strata so that for any two columns with m and n
        strata deposited, the rank of the most recent common ancestor can be
        determined with perfect certainty.

        Parameters
        ----------
        stratum_rank : int
            The number of strata that were deposited before the stratum under
            consideration for retention.
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.

        Returns
        -------
        bool
            True if the stratum should be retained, False otherwise.
        """

        return True

    def __eq__(
        self: 'StratumRetentionPredicatePerfectResolution',
        other: 'StratumRetentionPredicatePerfectResolution',
    ) -> bool:
        """Compare for value-wise equality."""

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def CalcNumStrataRetainedExact(
        self: 'StratumRetentionPredicatePerfectResolution',
        num_strata_deposited: int,
    ) -> int:
        """Exactly how many strata are retained after n deposted?"""

        return num_strata_deposited

    def CalcNumStrataRetainedUpperBound(
        self: 'StratumRetentionPredicatePerfectResolution',
        num_strata_deposited: int,
    ) -> int:
        """At most, how many strata are retained after n deposted? Inclusive."""

        return self.CalcNumStrataRetainedExact(
            num_strata_deposited=num_strata_deposited,
        )

    def CalcMrcaUncertaintyUpperBound(
        self: 'StratumRetentionPredicatePerfectResolution',
        *,
        first_num_strata_deposited: typing.Optional[int]=None,
        second_num_strata_deposited: typing.Optional[int]=None,
        actual_rank_of_mrca: typing.Optional[int]=None,
    ) -> int:
        """At most, how much uncertainty to estimate rank of MRCA? Inclusive."""

        return 0

    def CalcRankAtColumnIndex(
        self: 'StratumRetentionPredicatePerfectResolution',
        index: int,
        num_strata_deposited: typing.Optional[int]=None,
    ) -> int:
        """After n strata have been deposited, what will the rank of the
        stratum at column index k be?

        Enables a HereditaryStratigraphicColumn using this predicate to
        optimize away storage of rank annotations on strata. Takes into the
        account the possiblity for in-progress stratum depositions that haven't
        been reflected in num_strata_deposited.
        """

        return index



from copy import deepcopy
import random
import unittest

random.seed(1)

from hstrat import hstrat


class TestStratumRetentionPredicateNominalResolution(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_equality(self):
        assert (
            hstrat.StratumRetentionPredicateNominalResolution()
            == hstrat.StratumRetentionPredicateNominalResolution()
        )

        original = hstrat.StratumRetentionPredicateNominalResolution()
        copy = deepcopy(original)
        assert original == copy

    def test_behavior(self):
        for num_stratum_depositions_completed in range(100):
            for stratum_rank in range(0, num_stratum_depositions_completed):
                assert not hstrat.StratumRetentionPredicateNominalResolution()(
                    num_stratum_depositions_completed
                    =num_stratum_depositions_completed,
                    stratum_rank=stratum_rank,
                ) or stratum_rank in (0, num_stratum_depositions_completed)

    def test_space_complexity(self):
        predicate = hstrat.StratumRetentionPredicateNominalResolution()
        column = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_predicate=predicate,
        )

        for column_strata_deposited in range(100):
            for stratum_rank in range(0, column_strata_deposited):
                upper_bound = predicate.CalcNumStrataRetainedUpperBound(
                    num_strata_deposited=column.GetNumStrataDeposited(),
                )
                assert column.GetNumStrataRetained() <= upper_bound

    def _do_test_resolution(self, synchronous):
        predicate = hstrat.StratumRetentionPredicateNominalResolution()
        column = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_predicate=predicate,
        )

        population = [
            column.Clone()
            for __ in range(25)
        ]

        for generation in range(500):
            # subsample consecutive pairs in population
            for f, s in  zip(population, population[1:]):
                target_resolu = predicate.CalcMrcaUncertaintyUpperBound(
                    first_num_strata_deposited=f.GetNumStrataDeposited(),
                    second_num_strata_deposited=s.GetNumStrataDeposited(),
                )
                assert f.CalcRankOfMrcaUncertaintyWith(s) <= target_resolu
                assert f.CalcRanksSinceMrcaUncertaintyWith(s) <= target_resolu

            random.shuffle(population)
            for target in range(5):
                population[target] = population[-1].Clone()
            for individual in population:
                if synchronous or random.choice([True, False]):
                    individual.DepositStratum()

    def test_resolution(self):
        for synchronous in [True, False]:
            self._do_test_resolution(synchronous)

    def test_CalcRankAtColumnIndex(self):
        predicate = hstrat.StratumRetentionPredicateNominalResolution()
        column = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_predicate=predicate,
            initial_stratum_annotation=0,
        )

        for generation in range(1,501):
            for index in range(column.GetNumStrataRetained()):
                assert (
                    column.GetStratumAtColumnIndex(index).GetAnnotation()
                    == column.GetRankAtColumnIndex(index)
                )
            column.DepositStratum(annotation=generation)


if __name__ == '__main__':
    unittest.main()



from copy import deepcopy
import random
import unittest

from hstrat import hstrat

random.seed(1)


class TestHereditaryStratum(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_equality(self):
        def make_populated_bundle():
            return hstrat.HereditaryStratigraphicColumnBundle({
                'test' : hstrat.HereditaryStratigraphicColumn(
                ),
                'control' : hstrat.HereditaryStratigraphicColumn(
                    stratum_retention_condemner
                        =hstrat.StratumRetentionCondemnerPerfectResolution(),
                ),
            })

        assert make_populated_bundle() != make_populated_bundle()
        column1 = make_populated_bundle()
        column2 = column1
        assert column1 == column2
        assert column1 == deepcopy(column2)

        column3 = make_populated_bundle()
        assert column3 != column2

    def test_clone(self):
        def make_populated_bundle():
            return hstrat.HereditaryStratigraphicColumnBundle({
                'test' : hstrat.HereditaryStratigraphicColumn(
                ),
                'control' : hstrat.HereditaryStratigraphicColumn(
                    stratum_retention_condemner
                        =hstrat.StratumRetentionCondemnerPerfectResolution(),
                ),
            })

        column1 = make_populated_bundle()
        column2 = column1.Clone()
        column3 = column1.Clone()

        assert column1 == column2
        assert column2 == column3
        assert column1 == column3

        column1.DepositStratum()

        assert column1 != column2
        assert column2 == column3
        assert column1 != column3

    def test_forwarding_fallback(self):
        bundle1 = hstrat.HereditaryStratigraphicColumnBundle({
            'test' : hstrat.HereditaryStratigraphicColumn(
                stratum_retention_condemner
                    =hstrat.StratumRetentionCondemnerNominalResolution(),
                stratum_differentia_bit_width=1,
            ),
            'control' : hstrat.HereditaryStratigraphicColumn(
                stratum_retention_condemner
                    =hstrat.StratumRetentionCondemnerPerfectResolution(),
            ),
        })

        for __ in range(100): bundle1.DepositStratum()
        bundle2 = bundle1.Clone()
        for __ in range(100): bundle1.DepositStratum()
        for __ in range(100): bundle2.DepositStratum()

        res = bundle1.HasAnyCommonAncestorWith(bundle2)
        assert res['test'] is None
        assert res['control'] == True

        res = bundle1.HasAnyCommonAncestorWith(bundle2, confidence_level=0.49)
        assert res['test'] == True
        assert res['control'] == True

        res = bundle1.GetNumStrataRetained()
        assert \
            0 < res['test'] < res['control'] <= bundle1.GetNumStrataDeposited()

        assert bundle1._stratum_differentia_bit_width == {
            'test' : 1,
            'control' : 64,
        }

if __name__ == '__main__':
    unittest.main()



"""Functors that implement stratum retention policies by specifying
whether a stratum with deposition rank r should be retained within
a hereditary stratigraphic column after n strata have been
deposited."""


from .StratumRetentionPredicateDepthProportionalResolution \
    import StratumRetentionPredicateDepthProportionalResolution
from .StratumRetentionPredicateNominalResolution import StratumRetentionPredicateNominalResolution
from .StratumRetentionPredicatePerfectResolution import StratumRetentionPredicatePerfectResolution
from .StratumRetentionPredicateFixedResolution \
    import StratumRetentionPredicateFixedResolution
from .StratumRetentionPredicateRecencyProportionalResolution \
    import StratumRetentionPredicateRecencyProportionalResolution
from .StratumRetentionPredicateStochastic \
    import StratumRetentionPredicateStochastic

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    'StratumRetentionPredicateDepthProportionalResolution',
    'StratumRetentionPredicateFixedResolution',
    'StratumRetentionPredicatePerfectResolution',
    'StratumRetentionPredicateNominalResolution',
    'StratumRetentionPredicateRecencyProportionalResolution',
    'StratumRetentionPredicateStochastic',
]



from copy import deepcopy
import random
import unittest

random.seed(1)

from hstrat import hstrat


class TestStratumRetentionPredicatePerfectResolution(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_equality(self):
        assert (
            hstrat.StratumRetentionPredicatePerfectResolution()
            == hstrat.StratumRetentionPredicatePerfectResolution()
        )

        original = hstrat.StratumRetentionPredicatePerfectResolution()
        copy = deepcopy(original)
        assert original == copy

    def _do_test_resolution(self, synchronous):
        predicate = hstrat.StratumRetentionPredicatePerfectResolution()
        column = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_predicate=predicate,
        )

        population = [
            column.Clone()
            for __ in range(25)
        ]

        for generation in range(500):
            # subsample consecutive pairs in population
            for f, s in  zip(population, population[1:]):
                target_resolu = predicate.CalcMrcaUncertaintyUpperBound(
                    first_num_strata_deposited=f.GetNumStrataDeposited(),
                    second_num_strata_deposited=s.GetNumStrataDeposited(),
                )
                assert f.CalcRankOfMrcaUncertaintyWith(s) <= target_resolu
                assert f.CalcRanksSinceMrcaUncertaintyWith(s) <= target_resolu

            random.shuffle(population)
            for target in range(5):
                population[target] = population[-1].Clone()
            for individual in population:
                if synchronous or random.choice([True, False]):
                    individual.DepositStratum()

        def test_resolution(self):
            for synchronous in [True, False]:
                self._do_test_resolution(synchronous)

    def test_CalcRankAtColumnIndex(self):
        predicate = hstrat.StratumRetentionPredicatePerfectResolution()
        column = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_predicate=predicate,
            initial_stratum_annotation=0,
        )

        for generation in range(1,501):
            for index in range(column.GetNumStrataRetained()):
                assert (
                    column.GetStratumAtColumnIndex(index).GetAnnotation()
                    == column.GetRankAtColumnIndex(index)
                )
            column.DepositStratum(annotation=generation)


if __name__ == '__main__':
    unittest.main()



from copy import deepcopy
import random
import unittest

from hstrat import hstrat

random.seed(1)


class TestHereditaryStratum(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_deposition_rank(self):
        assert hstrat.HereditaryStratum(
            deposition_rank=42,
        ).GetDepositionRank() == 42

    def test_differentia_generation(self):
        original1 = hstrat.HereditaryStratum(
            deposition_rank=42,
        )
        copy1 = deepcopy(original1)
        original2 = hstrat.HereditaryStratum(
            deposition_rank=42,
        )

        assert original1 == copy1
        assert original1 != original2
        assert copy1 != original2

        assert original1.GetDifferentia() == copy1.GetDifferentia()
        assert original1.GetDifferentia() != original2.GetDifferentia()
        assert copy1.GetDifferentia() != original2.GetDifferentia()

    def test_equality1(self):
        assert hstrat.HereditaryStratum() != hstrat.HereditaryStratum()
        stratum1 = hstrat.HereditaryStratum()
        stratum2 = stratum1
        assert stratum1 == stratum2
        assert stratum1 == deepcopy(stratum2)

    def test_equality2(self):
        stratum_factory = lambda: hstrat.HereditaryStratum(deposition_rank=42)
        assert stratum_factory() != stratum_factory()
        stratum1 = stratum_factory()
        stratum2 = stratum1
        assert stratum1 == stratum2
        assert stratum1 == deepcopy(stratum2)


if __name__ == '__main__':
    unittest.main()



"""Functors that implement stratum retention policies by specifying
the set of strata ranks that should be pruned from a hereditary
stratigraphic column when the nth stratum is deposited."""

from .StratumRetentionCondemnerDepthProportionalResolution \
    import StratumRetentionCondemnerDepthProportionalResolution
from .StratumRetentionCondemnerFixedResolution \
    import StratumRetentionCondemnerFixedResolution
from .StratumRetentionCondemnerFromPredicate \
    import StratumRetentionCondemnerFromPredicate
from .StratumRetentionCondemnerPerfectResolution \
    import StratumRetentionCondemnerPerfectResolution
from .StratumRetentionCondemnerNominalResolution \
    import StratumRetentionCondemnerNominalResolution
from .StratumRetentionCondemnerRecencyProportionalResolution \
    import StratumRetentionCondemnerRecencyProportionalResolution

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    'StratumRetentionCondemnerDepthProportionalResolution',
    'StratumRetentionCondemnerFixedResolution',
    'StratumRetentionCondemnerFromPredicate',
    'StratumRetentionCondemnerPerfectResolution',
    'StratumRetentionCondemnerNominalResolution',
    'StratumRetentionCondemnerRecencyProportionalResolution',
]



from collections import deque
from copy import copy
from interval_search import binary_search
from iterpop import iterpop as ip
import itertools as it
import math
import operator
import opytional as opyt
import typing
import warnings

from .HereditaryStratum import HereditaryStratum
from .HereditaryStratumOrderedStoreList import HereditaryStratumOrderedStoreList

from .stratum_retention_condemners import StratumRetentionCondemnerFromPredicate
from .stratum_retention_condemners \
    import StratumRetentionCondemnerPerfectResolution


class HereditaryStratigraphicColumn:
    """Genetic annotation to enable phylogenetic inference among distributed
    digital evolution populations.

    Primary end-user facing interface for hstrat library. Should be bundled with
    digital genomes and propagated via the CloneDescendant method when passing
    those genomes from parent to offspring.

    Naming conventions are derived by analogy to Geological "Stratigraphy"
    (i.e., <https://en.wikipedia.org/wiki/Stratigraphy>). The "hereditary
    stratigraphy" system provided by this software works by associating an
    identifier, referred to as a "stratum," with each elapsed generation along
    a line of descent. This allows two "columns" to be aligned to detect the
    generation of their most recent common ancestor: strata before the MRCA
    will be identical and strata after will differ.

    Stratum retention policy and stratum differentia bit width can be configured
    to tune the time and space complexity of the column, trading-off with
    uncertainty induced on estimates of phylogenetic distance back to the most
    common recent ancestor of two columns.

    Arbitrary user-defined data can be associated with strata by optional
    argument to the CloneDescendant method. (Note that a first stratum is
    deposited during column initialization, so an optional annotation argument
    may also be provided then.)
    """

    # if True, strata will be constructed with deposition rank stored even if
    # the stratum retention condemner does not require it
    _always_store_rank_in_stratum: bool
    # how many bits wide of differentia should the deposted strata be
    # constructed with?
    _stratum_differentia_bit_width: int
    # counter tracking the number of strata deposited
    # incremented *after* a deposition and its coinciding purge are complete
    _num_strata_deposited: int
    # data structure storing retained strata
    _stratum_ordered_store: typing.Any
    # functor specifying stratum retention policy
    _stratum_retention_condemner: typing.Callable

    def __init__(
        self: 'HereditaryStratigraphicColumn',
        *,
        always_store_rank_in_stratum: bool=True,
        stratum_differentia_bit_width: int=64,
        initial_stratum_annotation: typing.Optional[typing.Any]=None,
        stratum_retention_condemner: typing.Callable=None,
        stratum_retention_predicate: typing.Callable=None,
        stratum_ordered_store_factory: typing.Callable
            =HereditaryStratumOrderedStoreList,
    ):
        """Initialize column to track a new line of descent.

        Deposits a first stratum, so GetNumStrataDeposited() will return 1 after
        initialization even though the user has not yet called DepositStratum().

        Parameters
        ----------
        always_store_rank_in_stratum : bool, optional
            Should the deposition rank be stored as a data member of generated
            strata, even if not strictly necessary?
        stratum_differentia_bit_width : int, optional
            The bit width of the generated differentia. Default 64, allowing
            for 2^64 distinct values.
        initial_stratum_annotation: any, optional
            Optional object to store as an annotation. Allows arbitrary user-
            provided to be associated with the first stratum deposition in the
            line of descent.
        stratum_retention_condemner : callable, optional
            Functor that implements stratum retention policy by specifying
            the set of strata ranks that should be pruned from a hereditary
            stratigraphic column when the nth stratum is deposited. Mutually
            exclusive with stratum_retention_predicate; only one should be
            specified.
        stratum_retention_predicate : callable, optional
            Functor that implements stratum retention policy by specifying
            whether a stratum with deposition rank r should be retained within
            a hereditary stratigraphic column after n strata have been
            deposited. Mutually exclusive with stratum_retention_condemner;
            only one should be specified.
        stratum_ordered_store_factory : callable, optional
            Callable to generate a container that implements the necessary
            interface to store strata within the column. Can be configured for
            performance reasons, but has no semantic effect. A type that can be
            default-constructed will suffice.

        Notes
        -----
        If no condemner or predicate functor specifying a stratum retention
        policy is provided, the perfect resolution policy where all strata are
        retained is used.
        """
        self._always_store_rank_in_stratum = always_store_rank_in_stratum
        self._stratum_differentia_bit_width = stratum_differentia_bit_width
        self._num_strata_deposited = 0
        self._stratum_ordered_store = stratum_ordered_store_factory()

        if None not in (
            stratum_retention_predicate,
            stratum_retention_condemner,
        ):
            raise ValueError(
                'Exactly one of `stratum_retention_condemner` '
                'and `stratum_retention_predicate` must be provided.'
            )
        else:
            self._stratum_retention_condemner = (
                StratumRetentionCondemnerFromPredicate(
                    stratum_retention_predicate,
                )
                    if (stratum_retention_predicate is not None)
                else stratum_retention_condemner
                    if stratum_retention_condemner is not None
                else StratumRetentionCondemnerPerfectResolution()
            )

        self.DepositStratum(annotation=initial_stratum_annotation)

    def __eq__(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
    ) -> bool:
        """Compare for value-wise equality."""

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def _ShouldOmitStratumDepositionRank(
        self: 'HereditaryStratigraphicColumn',
    ) -> bool:
        """Implementation detail to inspect configured stratum retention policy
        and manual override to decide whether deposition rank should be stored
        as a data member of generated strata.

        Note that strata are not required to be stored if the stratum retention
        policy allows for the rank of a stratum to be calcualted as a function
        of its position within the column and the number of strata deposited
        onto the column. However, it may be beneficial to store the stratum
        anyways for performance reasons if this calculation is expenxive.
        """

        can_omit_deposition_rank = hasattr(
            self._stratum_retention_condemner,
            'CalcRankAtColumnIndex',
        )
        return (
            can_omit_deposition_rank
            and not self._always_store_rank_in_stratum
        )

    def DepositStratum(
        self: 'HereditaryStratigraphicColumn',
        annotation: typing.Optional[typing.Any]=None,
    ) -> None:
        """Elapse a generation.

        Parameters
        ----------
        annotation: any, optional
            Optional object to store as an annotation. Allows arbitrary user-
            provided to be associated with this stratum deposition in the
            line of descent.
        """

        new_stratum = HereditaryStratum(
            annotation=annotation,
            deposition_rank=(
                # don't store deposition rank if we know how to calcualte it
                # from stratum's position in column
                None
                if self._ShouldOmitStratumDepositionRank()
                else self._num_strata_deposited
            ),
            differentia_bit_width=self._stratum_differentia_bit_width,
        )
        self._stratum_ordered_store.DepositStratum(
            rank=self._num_strata_deposited,
            stratum=new_stratum,
        )
        self._PurgeColumn()
        self._num_strata_deposited += 1

    def _PurgeColumn(self: 'HereditaryStratigraphicColumn') -> None:
        """Implementation detail to discard stored strata according to the
        configured stratum retention policy.

        Called after a new stratum has been appended to the column's store but
        before it is considered fully deposited (i.e., it is reflected in the
        column's internal deposition counter).
        """

        condemned_ranks = self._stratum_retention_condemner(
            retained_ranks=self.GetRetainedRanks(),
            num_stratum_depositions_completed=self.GetNumStrataDeposited(),
        )
        self._stratum_ordered_store.DelRanks(
            ranks=condemned_ranks,
            get_column_index_of_rank=self.GetColumnIndexOfRank,
        )

    def GetRetainedRanks(
        self: 'HereditaryStratigraphicColumn',
    ) -> typing.Iterator[int]:
        """Get an iterator over deposition ranks of strata stored in the column.

        Order of iteration should not be considered guaranteed. The store may
        be altered during iteration without iterator invalidation, although
        subsequent updates will not be reflected in the iterator.
        """

        if self._ShouldOmitStratumDepositionRank():
            for idx in range(self.GetNumStrataRetained()):
                yield self.GetRankAtColumnIndex(idx)
        else:
            yield from self._stratum_ordered_store.GetRetainedRanks()

    def GetNumStrataRetained(self: 'HereditaryStratigraphicColumn') -> int:
        """How many strata are currently stored within the column?

        May be fewer than the number of strata deposited if strata have been
        discarded as part of the configured stratum retention policy.
        """

        return self._stratum_ordered_store.GetNumStrataRetained()

    def GetNumStrataDeposited(self: 'HereditaryStratigraphicColumn') -> int:
        """How many strata have been depostited on the column?

        Note that a first stratum is deposited on the column during
        initialization."""

        return self._num_strata_deposited

    def GetStratumAtColumnIndex(
        self: 'HereditaryStratigraphicColumn',
        index: int,
    ) -> HereditaryStratum:
        """Get the stratum positioned at index i among retained strata.

        Index order is from most ancient (index 0) to most recent.
        """

        return self._stratum_ordered_store.GetStratumAtColumnIndex(
            index,
            get_rank_at_column_index=(
                self.GetRankAtColumnIndex
                if self._ShouldOmitStratumDepositionRank()
                else None
            ),
        )

    def GetRankAtColumnIndex(
        self: 'HereditaryStratigraphicColumn',
        index: int,
    ) -> int:
        """What is the deposition rank of the stratum positioned at index i
        among retained strata?

        Index order is from most ancient (index 0) to most recent.
        """

        if self._ShouldOmitStratumDepositionRank():
            return self._stratum_retention_condemner.CalcRankAtColumnIndex(
                index=index,
                num_strata_deposited=self.GetNumStrataDeposited(),
            )
        else:
            # fall back to store lookup
            return self._stratum_ordered_store.GetRankAtColumnIndex(index)

    def GetColumnIndexOfRank(
        self: 'HereditaryStratumOrderedStoreList',
        rank: int,
    ) -> typing.Optional[int]:
        """What is the index position within retained strata of the stratum
        deposited at rank r?

        Returns None if no stratum with rank r is present within the store.
        """

        if self._ShouldOmitStratumDepositionRank():
            assert self.GetNumStrataRetained()
            res_idx = binary_search(
                lambda idx: self.GetRankAtColumnIndex(idx) >= rank,
                0,
                self.GetNumStrataRetained() - 1,
            )
            if res_idx is None:
                return None
            elif self.GetRankAtColumnIndex(res_idx) == rank:
                return res_idx
            else:
                return None
        else:
            # fall back to store lookup
            return self._stratum_ordered_store.GetColumnIndexOfRank(rank=rank)

    def GetNumDiscardedStrata(
        self: 'HereditaryStratigraphicColumn',
    ) -> int:
        """How many strata have been discarded by the configured column
        retention policy?"""

        return self.GetNumStrataDeposited() - self.GetNumStrataRetained()

    def HasDiscardedStrata(
        self: 'HereditaryStratigraphicColumn',
    ) -> bool:
        """Have any strata have been discarded by the configured column
        retention policy?"""

        return self.GetNumDiscardedStrata() > 0

    def CalcProbabilityDifferentiaCollision(
        self: 'HereditaryStratigraphicColumn',
    ) -> float:
        """What is the probability of two randomly-differentiated differentia
        being identical by coincidence?"""

        return 1.0 / 2**self._stratum_differentia_bit_width

    def CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
        self: 'HereditaryStratigraphicColumn',
        *,
        significance_level: float,
    ) -> float:
        """How many differentia collisions are required to reject the null
        hypothesis that columns do not share common ancestry at those ranks at
        significance level significance_level?"""

        assert 0.0 <= significance_level <= 1.0

        log_base = self.CalcProbabilityDifferentiaCollision()
        return int(math.ceil(
            math.log(significance_level, log_base)
        ))

    def CalcDefinitiveMaxRankOfLastRetainedCommonalityWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
    ) -> typing.Optional[int]:
        """At most, how many depositions elapsed along the columns' lines of
        descent before the last matching strata at the same rank between
        self and other?

        Returns
        -------
        int, optional
            The number of depositions elapsed or None if no common ancestor is
            shared between the columns.
        """

        confidence_level = 0.49
        assert self.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1.0 - confidence_level,
        ) == 1
        return self.CalcRankOfLastRetainedCommonalityWith(
            other,
            confidence_level=confidence_level,
        )

    def CalcRankOfLastRetainedCommonalityWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float=0.95,
    ) -> typing.Optional[int]:
        """How many depositions elapsed along the columns' lines of
        descent before the last matching strata at the same rank between
        self and other?

        Parameters
        ----------
        confidence_level : float, optional
            With what probability should the true rank of the last commonality
            with other fall at or after the returned rank? Default 0.95.

        Returns
        -------
        int, optional
            The number of depositions elapsed or None if no common ancestor is
            shared between the columns.

        Notes
        -----
        The true rank of the last commonality with other is guaranteed to never
        be after the returned rank when confidence_level < 0.5.
        """

        assert 0.0 <= confidence_level <= 1.0

        if (
            self.HasDiscardedStrata()
            or other.HasDiscardedStrata()
            # for performance reasons
            # only binary search stores that support random access
            or not isinstance(
                self._stratum_ordered_store,
                HereditaryStratumOrderedStoreList
            )
            or not isinstance(
                other._stratum_ordered_store,
                HereditaryStratumOrderedStoreList
            )
            # binary search currently requires no spurious collisions
            or self._stratum_differentia_bit_width < 64
        ):
            return self._do_generic_CalcRankOfLastRetainedCommonalityWith(
                other,
                confidence_level=confidence_level,
            )
        else:
            return self._do_binary_search_CalcRankOfLastRetainedCommonalityWith(
                other,
                confidence_level=confidence_level,
            )

    def _do_binary_search_CalcRankOfLastRetainedCommonalityWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float,
    ) -> typing.Optional[int]:
        """Implementation detail with optimized implementation for specialized
        case where both self and other use the perfect resolution stratum
        retention policy."""

        # both must have (effectively) used the perfect resolution policy
        assert not self.HasDiscardedStrata() and not other.HasDiscardedStrata()

        lower_bound = 0
        upper_bound = min([
            self.GetNumStrataDeposited() - 1,
            other.GetNumStrataDeposited() - 1,
        ])
        assert lower_bound <= upper_bound
        rank_at = lambda which, idx: which.GetRankAtColumnIndex(idx)
        differentia_at = lambda which, idx: \
                which.GetStratumAtColumnIndex(idx).GetDifferentia()
        # todo: how to account for possible spurious differentiae collisions?
        predicate = lambda idx: \
            differentia_at(self, idx) != differentia_at(other, idx)

        first_disparite_idx = binary_search(
            predicate,
            lower_bound,
            upper_bound,
        )

        collision_implausibility_threshold = self.\
            CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=1.0 - confidence_level,
        )
        assert collision_implausibility_threshold > 0
        if first_disparite_idx is None:
            # no disparate strata found
            # fall back to _do_generic_CalcRankOfLastRetainedCommonalityWith to handle
            # proper bookkeeping in this case while skipping most of the search
            return self._do_generic_CalcRankOfLastRetainedCommonalityWith(
                other,
                self_start_idx=upper_bound,
                other_start_idx=upper_bound,
                confidence_level=confidence_level,
            )
        elif first_disparite_idx >= collision_implausibility_threshold:
            # disparate strata found, following some common strata
            # ...discount collision_implausibility_threshold - 1 common strata
            # as potential spurious differentia collisions
            # ... must also subtract 1 (canceling out -1 above) to account for
            # moving from disparite stratum to preceding common stratum
            last_common_idx \
                = first_disparite_idx - collision_implausibility_threshold
            last_common_rank = self.GetRankAtColumnIndex(
                last_common_idx,
            )
            assert last_common_idx == last_common_rank
            return last_common_rank
        else:
            # no common strata between self and other
            # or not enough common strata to discount the possibility all
            # are spurious collisions with respect to the given confidence
            # level; conservatively conclude there is no common ancestor
            return None

    def _do_generic_CalcRankOfLastRetainedCommonalityWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        *,
        self_start_idx: int=0,
        other_start_idx: int=0,
        confidence_level: float,
    ) -> typing.Optional[int]:
        """Implementation detail with general-case implementation."""

        # helper setup
        self_iter = self._stratum_ordered_store.IterRankDifferentia(
            get_rank_at_column_index=self.GetRankAtColumnIndex,
            start_column_index=self_start_idx,
        )
        other_iter = other._stratum_ordered_store.IterRankDifferentia(
            get_rank_at_column_index=other.GetRankAtColumnIndex,
            start_column_index=other_start_idx,
        )
        self_cur_rank, self_cur_differentia = next(self_iter)
        other_cur_rank, other_cur_differentia = next(other_iter)

        # we need to keep track of enough ranks of last-seen common strata so
        # that we can discount this many (minus 1) as potentially occuring due
        # to spurious differentia collisions
        collision_implausibility_threshold = self.\
            CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=1.0 - confidence_level,
        )
        assert collision_implausibility_threshold > 0
        # holds up to n last-seen ranks with common strata,
        # with the newest last-seen rank at the front (index 0)
        # and the up to nth last-seen rank at the back (index -1)
        preceding_common_strata_ranks \
            = deque([], collision_implausibility_threshold)
        # a.k.a.
        # while (
        #     self_column_idx < self.GetNumStrataRetained()
        #     and other_column_idx < other.GetNumStrataRetained()
        # ):
        try:
            while True:
                if (self_cur_rank == other_cur_rank):
                    # strata at same rank can be compared
                    if (self_cur_differentia == other_cur_differentia):
                        # matching differentiae at the same rank,
                        # store rank and keep searching for mismatch
                        preceding_common_strata_ranks.appendleft(self_cur_rank)
                        # advance self
                        self_cur_rank, self_cur_differentia = next(self_iter)
                        # advance other
                        other_cur_rank, other_cur_differentia = next(other_iter)
                    else:
                        # mismatching differentiae at the same rank
                        # a.k.a. break
                        raise StopIteration
                elif self_cur_rank < other_cur_rank:
                    # current stratum on self column older than on other column
                    # advance to next-newer stratum on self column
                    self_cur_rank, self_cur_differentia = next(self_iter)
                elif self_cur_rank > other_cur_rank:
                    # current stratum on other column older than on self column
                    # advance to next-newer stratum on other column
                    other_cur_rank, other_cur_differentia = next(other_iter)
        except StopIteration:
            try:
                # discount collision_implausibility_threshold - 1 common strata
                # as potential spurious differentia collisions
                return preceding_common_strata_ranks[
                    collision_implausibility_threshold - 1
                ]
            except IndexError:
                # not enough common strata to discount the possibility all
                # are spurious collisions with respect to the given confidence
                # level; conservatively conclude there is no common ancestor
                return None

    def GetNthCommonRankWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        n: int,
    ) -> typing.Optional[int]:
        """Return the nth rank retained by both columns.

        Zero indexed. Returns None if n + 1 common ranks do not exist between
        self and other.
        """

        assert n >= 0

        # helper setup
        self_iter = self._stratum_ordered_store.IterRankDifferentia(
            get_rank_at_column_index=self.GetRankAtColumnIndex,
        )
        other_iter = other._stratum_ordered_store.IterRankDifferentia(
            get_rank_at_column_index=other.GetRankAtColumnIndex,
        )
        self_cur_rank, self_cur_differentia = next(self_iter)
        other_cur_rank, other_cur_differentia = next(other_iter)

        try:
            while True:
                if self_cur_rank == other_cur_rank:
                    # strata at common rank
                    if n == 0: return self_cur_rank

                    n -= 1
                    # advance self
                    self_cur_rank, self_cur_differentia = next(self_iter)
                    # advance other
                    other_cur_rank, other_cur_differentia = next(other_iter)
                elif self_cur_rank < other_cur_rank:
                    # current stratum on self column older than on other column
                    # advance to next-newer stratum on self column
                    self_cur_rank, self_cur_differentia = next(self_iter)
                elif self_cur_rank > other_cur_rank:
                    # current stratum on other column older than on self column
                    # advance to next-newer stratum on other column
                    other_cur_rank, other_cur_differentia = next(other_iter)

        except StopIteration:
            return None


    def CalcDefinitiveMaxRankOfFirstRetainedDisparityWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
    ) -> typing.Optional[int]:
        """At most, how many depositions elapsed along the columns'
        lines of descent before the first mismatching strata at the same rank
        between self and other?

        Returns
        -------
        int, optional
            The number of depositions elapsed or None if no disparity (i.e.,
            both columns have same number of strata deposited and the most
            recent stratum is common between self and other).

        Notes
        -----
        If no mismatching strata are found but self and other have different
        numbers of strata deposited, this method returns one greater than the
        lesser of the columns' deposition counts.
        """

        confidence_level = 0.49
        assert self.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1.0 - confidence_level,
        ) == 1
        return self.CalcRankOfFirstRetainedDisparityWith(
            other,
            confidence_level=confidence_level,
        )

    def CalcRankOfFirstRetainedDisparityWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float=0.95,
    ) -> typing.Optional[int]:
        """How many depositions elapsed along the columns' lines of
        descent before the first mismatching strata at the same rank between
        self and other?

        Parameters
        ----------
        confidence_level : float, optional
            With what probability should the true rank of the first disparity
            with other fall at or after the returned rank? Default 0.95.

        Returns
        -------
        int, optional
            The number of depositions elapsed or None if no disparity (i.e.,
            both columns have same number of strata deposited and the most
            recent stratum is common between self and other).

        Notes
        -----
        If no mismatching strata are found but self and other have different
        numbers of strata deposited, this method returns one greater than the
        lesser of the columns' deposition counts.

        The true rank of the first disparity with other is guaranteed to never
        be after the returned rank when confidence_level < 0.5.

        If the differentia width and confidence level are configured such that
        one or more suprious differentia collisions is plausible, this method
        will never return None.

        Consider two columns that compare identical at all common ranks. If
        these columns have few enough common ranks that all these collisions
        could plausibly be spurious then there is not yet enough to support
        common lineage between the columns so rank 0 must be selected as the
        estimated site of first retained disparity. If these columns have
        enough common ranks that all collisions could not plausibly be spurious
        then the largest rank that could not conceivably be spurious must be
        selected (and we are guaranteed such a rank exists). It is only when
        not even one collision could plausibly be spurious that strong enough
        evidence exists to conclude that there is no disparity.

        For example, the method may return None with 64-bit differentia and 95%
        confidence level but will never return None with 1-bit differentia and
        95% confidence level.
        """

        assert 0.0 <= confidence_level <= 1.0

        if (
            self.HasDiscardedStrata()
            or other.HasDiscardedStrata()
            # for performance reasons
            # only apply binary search to stores that support random access
            or not isinstance(
                self._stratum_ordered_store,
                HereditaryStratumOrderedStoreList
            )
            or not isinstance(
                other._stratum_ordered_store,
                HereditaryStratumOrderedStoreList
            )
            # binary search currently requires no spurious collisions
            or self._stratum_differentia_bit_width < 64
        ):
            return self._do_generic_CalcRankOfFirstRetainedDisparityWith(
                other,
                confidence_level=confidence_level,
            )
        else:
            return self._do_binary_search_CalcRankOfFirstRetainedDisparityWith(
                other,
                confidence_level=confidence_level,
            )

    def _do_binary_search_CalcRankOfFirstRetainedDisparityWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float,
    ) -> typing.Optional[int]:
        """Implementation detail with optimized implementation for specialized
        case where both self and other use the perfect resolution stratum
        retention policy."""

        # both must have (effectively) used the perfect resolution policy
        assert not self.HasDiscardedStrata() and not other.HasDiscardedStrata()

        lower_bound = 0
        upper_bound = min([
            self.GetNumStrataDeposited() - 1,
            other.GetNumStrataDeposited() - 1,
        ])
        assert lower_bound <= upper_bound
        rank_at = lambda which, idx: which.GetRankAtColumnIndex(idx)
        differentia_at = lambda which, idx: \
                which.GetStratumAtColumnIndex(idx).GetDifferentia()
        # todo: how to account for possible spurious differentiae collisions?
        predicate = lambda idx: \
            differentia_at(self, idx) != differentia_at(other, idx)

        first_disparite_idx = binary_search(
            predicate,
            lower_bound,
            upper_bound,
        )

        if first_disparite_idx is not None:
            # disparate strata found
            collision_plausibility_threshold = self.\
                CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                    significance_level=1.0 - confidence_level,
            ) - 1
            # discount collision_implausibility_threshold - 1 common
            # ranks due to potential spurious differentia collisions;
            # if not enough common ranks are available we still know
            # *definitively* that a disparity occured (because we
            # observed disparite strata at the same rank); so, make the
            # conservative assumption that the disparity occured as far
            # back as possible (rank 0)
            spurious_collision_corrected_idx = max(
                first_disparite_idx - collision_plausibility_threshold,
                0,
            )
            first_disparite_rank = self.GetRankAtColumnIndex(
                spurious_collision_corrected_idx
            )
            return first_disparite_rank
        else:
            # no disparate strata found
            # fall back to _do_generic_CalcRankOfFirstRetainedDisparityWith to
            # handle proper bookkeeping in this case while skipping most of the
            # search
            return self._do_generic_CalcRankOfFirstRetainedDisparityWith(
                other,
                self_start_idx=upper_bound,
                other_start_idx=upper_bound,
                confidence_level=confidence_level,
            )

    def _do_generic_CalcRankOfFirstRetainedDisparityWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        *,
        self_start_idx: int=0,
        other_start_idx: int=0,
        confidence_level: float,
    ) -> typing.Optional[int]:
        """Implementation detail with general-case implementation."""

        # helper setup
        self_iter = self._stratum_ordered_store.IterRankDifferentia(
            get_rank_at_column_index=self.GetRankAtColumnIndex,
            start_column_index=self_start_idx,
        )
        other_iter = other._stratum_ordered_store.IterRankDifferentia(
            get_rank_at_column_index=other.GetRankAtColumnIndex,
            start_column_index=other_start_idx,
        )
        self_cur_rank, self_cur_differentia = next(self_iter)
        other_cur_rank, other_cur_differentia = next(other_iter)
        self_prev_rank: int
        other_prev_rank: int

        def advance_self():
            nonlocal self_prev_rank, self_cur_rank, \
                self_cur_differentia, self_iter
            try:
                self_prev_rank = self_cur_rank
                self_cur_rank, self_cur_differentia = next(self_iter)
            except StopIteration:
                self_iter = None

        def advance_other():
            nonlocal other_prev_rank, other_cur_rank, \
                other_cur_differentia, other_iter
            try:
                other_prev_rank = other_cur_rank
                other_cur_rank, other_cur_differentia = next(other_iter)
            except StopIteration:
                other_iter = None


        # we need to keep track of enough last-seen common ranks so that we
        # can discount this many (minus 1) as potentially occuring due to
        # spurious differentia collisions
        collision_implausibility_threshold = self.\
            CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=1.0 - confidence_level,
        )
        assert collision_implausibility_threshold > 0
        # holds up to n last-seen common ranks,
        # with the newest last-seen rank at the front (index 0)
        # and the up to nth last-seen rank at the back (index -1)
        preceding_common_ranks = deque([], collision_implausibility_threshold)
        # a.k.a.
        # while (
        #     self_column_idx < self.GetNumStrataRetained()
        #     and other_column_idx < other.GetNumStrataRetained()
        # ):
        while self_iter is not None and other_iter is not None:
            if self_cur_rank == other_cur_rank:
                # strata at same rank can be compared
                if self_cur_differentia == other_cur_differentia:
                    preceding_common_ranks.appendleft(self_cur_rank)
                    # matching differentiae at the same rank,
                    # keep searching for mismatch
                    # advance self and other
                    # must ensure both advance, even if one stops iteration
                    advance_self()
                    advance_other()
                else:
                    # mismatching differentiae at the same rank
                    preceding_common_ranks.appendleft(self_cur_rank)
                    assert 0 <= self_cur_rank < self.GetNumStrataDeposited()

                    # discount collision_implausibility_threshold - 1 common
                    # ranks due to potential spurious differentia collisions;
                    # if not enough common ranks are available we still know
                    # *definitively* that a disparity occured (because we
                    # observed disparite strata at the same rank); so, make the
                    # conservative assumption that the disparity occured as far
                    # back as possible (the oldest up to nth last-seen common
                    # rank, at the back of the deque)
                    return preceding_common_ranks[-1]
            elif self_cur_rank < other_cur_rank:
                # current stratum on self column older than on other column
                # advance to next-newer stratum on self column
                advance_self()
            elif self_cur_rank > other_cur_rank:
                # current stratum on other column older than on self column
                # advance to next-newer stratum on other column
                advance_other()

        if self_iter is not None:
            # although no mismatching strata found between self and other
            # self has strata ranks beyond the newest found in other
            # conservatively assume mismatch will be with next rank of other
            assert other_iter is None
            preceding_common_ranks.appendleft(other_prev_rank + 1)
            res = preceding_common_ranks[-1]
            assert 0 <= res <= self.GetNumStrataDeposited()
            assert 0 <= res <= other.GetNumStrataDeposited()
            return res
        elif other_iter is not None:
            # although no mismatching strata found between other and self
            # other has strata ranks beyond the newest found in self
            # conservatively assume mismatch will be with next rank
            assert self_iter is None
            preceding_common_ranks.appendleft(self_prev_rank + 1)
            res = preceding_common_ranks[-1]
            assert 0 <= res <= self.GetNumStrataDeposited()
            assert 0 <= res <= other.GetNumStrataDeposited()
            return res
        else:
            # no disparate strata found
            # and self and other have the same newest rank
            assert self_iter is None and other_iter is None
            preceding_common_ranks.appendleft(None)
            return preceding_common_ranks[-1]

    def CalcRankOfMrcaBoundsWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float=0.95,
    ) -> typing.Optional[typing.Tuple[int, int]]:
        """Calculate bounds on estimate for the number of depositions elapsed
        along the line of descent before the most recent common ancestor with
        other.

        Parameters
        ----------
        confidence_level : float, optional
            Bounds must capture what probability of containing the true rank of
            the MRCA? Default 0.95.

        Returns
        -------
        (int, int), optional
            Inclusive lower and then exclusive upper bound on estimate or None
            if no common ancestor between self and other can be resolved with
            sufficient confidence. (Sufficient confidence depends on
            bound_type.)

        See Also
        --------
        CalcRankOfMrcaUncertaintyWith :
            Wrapper to report uncertainty of calculated bounds.
        CalcRankOfEarliestDetectableMrcaWith :
            Could any MRCA be detected between self and other? What is the rank
            of the earliest MRCA that could be reliably detected?

        Notes
        -----
        The true rank of the MRCA is guaranteed to never fall above the bounds but may fall below.

        An alternate approach could be to construct the bounds such that the
        true rank of the MRCA will fall above or below the bounds with equal
        probability. This would involve setting the confidence level for
        calculating the first disparity with other to significance_level/2 and
        the confidence level for calculaing the last comonality with other to
        1 - significance_level/2. This means the confidence level applied to
        calculating the first disparity with other would always be <= 0.5.
        However, shifting the calculated first disparity with other below
        the definitive max first retained disparity requires confidence level
        >= 0.5. So, in practice such a symmetric approach would only result in
        the lower bound being shifted downward. For this reason, it is no longer
        provided as an option.

        In the absence of evidence to the contrary (i.e., more common
        strata than spurious differentia collisions alone could plausibly
        cause), this method assumes no common ancestry between self and other,
        returning None. This means that if few enough common ranks are shared
        between self and other (and the differentia bit with is small enough),
        it may not be possible to detect any common ancestry after accounting
        for the possibility of spurious differentia collisions (even if common
        ancestry did exist). So, calls to this method would always return None.
        Likewise, MRCAs at very early ranks may not be able to be reliably
        detected due to insufficient evidence. This can lead to cases where
        columns with true common ancestry have MRCA bounds estimated as None at
        much higher than the expected failure rate at the given confidence
        level. Note that with sufficient differentia bit width (i.e., so that
        even one collision is implausible at the given confidence level) this
        issue does not occur. Use CalcRankOfEarliestDetectableMrcaWith to
        determine the earliest rank at which an MRCA could be reliably detected
        between self and other.
        """

        assert 0.0 <= confidence_level <= 1.0

        if self.CalcRankOfEarliestDetectableMrcaWith(
            other,
            confidence_level=confidence_level,
        ) is None:
            warnings.warn(
                'Insufficient common ranks between columns to detect common ' 'ancestry at given confidence level.'
            )

        if self.HasAnyCommonAncestorWith(
            other,
            confidence_level=confidence_level,
        ):
            first_disparity \
                = self.CalcDefinitiveMaxRankOfFirstRetainedDisparityWith(
                        other,
                    )
            if first_disparity is None:
                num_self_deposited = self.GetNumStrataDeposited()
                num_other_deposited = other.GetNumStrataDeposited()
                assert num_self_deposited == num_other_deposited
            last_commonality = self.CalcRankOfLastRetainedCommonalityWith(
                other,
                confidence_level=confidence_level,
            )
            assert last_commonality is not None
            return (
                last_commonality,
                opyt.or_value(first_disparity, self.GetNumStrataDeposited()),
            )
        else: return None

    def CalcRankOfMrcaUncertaintyWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float=0.95,
    ) -> typing.Optional[int]:
        """Calculate uncertainty of estimate for the number of depositions
        elapsed along the line of descent before the most common recent
        ancestor with other.

        Returns 0 if no common ancestor between self and other can be resolved
        with sufficient confidence. If insufficient common ranks between self
        and other are available to resolve any common ancestor, returns None.

        See Also
        --------
        CalcRankOfMrcaBoundsWith :
            Calculates bound whose uncertainty this method reports. See the
            corresponding docstring for explanation of parameters.
        """

        if self.CalcRankOfEarliestDetectableMrcaWith(
            other,
            confidence_level=confidence_level,
        ) is None:
            return None

        bounds = self.CalcRankOfMrcaBoundsWith(
            other,
            confidence_level=confidence_level,
        )
        return 0 if bounds is None else abs(operator.sub(*bounds)) - 1

    def CalcRankOfEarliestDetectableMrcaWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float=0.95,
    ) -> typing.Optional[int]:
        """What is the earliest possible rank a MRCA between self and other
        could be reliably detected at?

        Even if a true MRCA of self and other exists, if it occured earlier
        than the rank calculated here it could not be reliably detected with
        sufficient confidence after accounting for the possibility of spurious
        differentia collisions. (Although subsequent spurious differentia
        collisions after the true MRCA of self and other could lead to MRCA
        detection at such a rank.)

        Returns None if insufficient common ranks exist between self and other
        to ever conclude at the given confidence level the existance of any
        common ancestry between self and other (even if all strata at common
        ranks had equivalent differentiae).
        """

        num_required_common_ranks = self.\
            CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=1 - confidence_level,
            )
        # zero-indexed
        return self.GetNthCommonRankWith(other, num_required_common_ranks - 1)

    def CalcDefinitiveMinRanksSinceLastRetainedCommonalityWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
    ) -> typing.Optional[int]:
        """At least, how many depositions have elapsed along this column's line
        of descent since the last matching strata at the same rank between self
        and other?

        Returns None if no common ancestor between self and other can be
        resolved with absolute confidence.
        """

        confidence_level = 0.49
        assert self.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1.0 - confidence_level,
        ) == 1
        return self.CalcRanksSinceLastRetainedCommonalityWith(
            self,
            confidence_level=confidence_level,
        )

    def CalcRanksSinceLastRetainedCommonalityWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float=0.95,
    ) -> typing.Optional[int]:
        """How many depositions have elapsed along this column's line of
        descent since the las matching strata at the same rank between self and
        other?

        Returns None if no common ancestor is shared between self and other.

        Parameters
        ----------
        confidence_level : float, optional
            With what probability should the true number of ranks since the
            last commonality with other be less than the calculated estimate?
            Default 0.95.

        Notes
        -----
        If confidence_level < 0.5, then the true number of ranks since the last
        commonality with other is guaranteed greater than or equal to the
        calculated estimate.
        """

        assert 0.0 <= confidence_level <= 1.0

        last_common_rank = self.CalcRankOfLastRetainedCommonalityWith(
            other,
            confidence_level=confidence_level,
        )
        if last_common_rank is None: return None
        else:
            assert self.GetNumStrataDeposited()
            res = self.GetNumStrataDeposited() - 1 - last_common_rank
            assert 0 <= res < self.GetNumStrataDeposited()
            return res

    def CalcDefinitiveMinRanksSinceFirstRetainedDisparityWith(
            self: 'HereditaryStratigraphicColumn',
            other: 'HereditaryStratigraphicColumn',
    ) -> typing.Optional[int]:
        """At least, how many depositions have elapsed along this column's line
        of descent since the first mismatching strata at the same rank between
        self and other?

        Returns None if no disparity found (i.e., both columns have same number
        of strata deposited and the most recent stratum is common between self
        and other).
        """

        confidence_level = 0.49
        assert self.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1.0 - confidence_level,
        ) == 1
        return self.CalcRanksSinceFirstRetainedDisparityWith(
            other,
            confidence_level=confidence_level,
        )

    def CalcRanksSinceFirstRetainedDisparityWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float=0.95,
    ) -> typing.Optional[int]:
        """How many depositions have elapsed along this column's line of
        descent since the first mismatching strata at the same rank between
        self and other?

        Returns
        -------
        int, optional
            The number of depositions elapsed or None if no disparity (i.e.,
            both columns have same number of strata deposited and the most
            recent stratum is common between self and other).

        Parameters
        ----------
        confidence_level : float, optional
            With what probability should the true number of ranks since the
            first disparity be less than or equal to the returned estimate?
            Default 0.95.

        Notes
        -----
        Returns -1 if self and other share no mismatching strata at common
        ranks but other has more strata deposited then self.

        The true number of ranks since the first disparity with other is
        guaranteed strictly less than or equal to the returned estimate when
        confidence_level < 0.5.
        """

        first_disparate_rank = self.CalcRankOfFirstRetainedDisparityWith(
            other,
            confidence_level=confidence_level,
        )
        if first_disparate_rank is None: return None
        else:
            assert self.GetNumStrataDeposited()
            res = self.GetNumStrataDeposited() - 1 - first_disparate_rank
            assert -1 <= res < self.GetNumStrataDeposited()
            return res

    def CalcRanksSinceMrcaBoundsWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float=0.95,
    ) -> typing.Optional[typing.Tuple[int, int]]:
        """Calculate bounds on estimate for the number of depositions elapsed
        along this column's line of descent since the most recent common
        ancestor with other.

        Parameters
        ----------
        confidence_level : float, optional
            With what probability should the true rank of the MRCA fall
            within the calculated bounds? Default 0.95.

        Returns
        -------
        (int, int), optional
            Inclusive lower bound and then exclusive upper bound on estimate or
            None if no common ancestor between self and other can be resolved
            with sufficient confidence. (Sufficient confidence depends on
            bound_type.)

        See Also
        --------
        CalcRanksSinceMrcaUncertaintyWith :
            Wrapper to report uncertainty of calculated bounds.
        CalcRanksSinceEarliestDetectableMrcaWith :
            Could any MRCA be detected between self and other? How many ranks
            have elapsed since the earliest MRCA that could be reliably
            detected?

        Notes
        -----
        The true number of ranks since the MRCA is guaranteed to never fall below the bounds but may fall above.

        An alternate approach could be to construct the bounds such that the
        true number of ranks since the MRCA will fall above or below the bounds
        with equal probability. This would involve setting the confidence level
        for calculating the first disparity with other to significance_level/2
        and the confidence level for calculaing the last comonality with other
        to 1 - significance_level/2. This means the confidence level applied to
        calculating the first disparity with other would always be <= 0.5.
        However, shifting the calculated first disparity with other above
        the definitive min first retained disparity requires confidence level
        >= 0.5. So, in practice such a symmetric approach would only result in
        the upper bound being shifted upward. For this reason, it is no longer
        provided as an option.

        In the absence of evidence to the contrary (i.e., more common
        strata than spurious differentia collisions alone could plausibly
        cause), this method assumes no common ancestry between self and other,
        returning None. This means that if few enough common ranks are shared
        between self and other (and the differentia bit with is small enough),
        it may not be possible to detect any common ancestry after accounting
        for the possibility of spurious differentia collisions (even if common
        ancestry did exist). So, calls to this method would always return None.
        Likewise, MRCAs at very early ranks may not be able to be reliably
        detected due to insufficient evidence. This can lead to cases where
        columns with true common ancestry have MRCA bounds estimated as None at
        much higher than the expected failure rate at the given confidence
        level. Note that with sufficient differentia bit width (i.e., so that
        even one collision is implausible at the given confidence level) this
        issue does not occur. Use CalcRanksSinceEarliestDetectableMrcaWith to
        determine the earliest rank at which an MRCA could be reliably detected
        between self and other.
        """

        assert 0.0 <= confidence_level <= 1.0

        if self.CalcRankOfEarliestDetectableMrcaWith(
            other,
            confidence_level=confidence_level,
        ) is None:
            warnings.warn(
                'Insufficient common ranks between columns to detect common ' 'ancestry at given confidence level.'
            )

        if self.HasAnyCommonAncestorWith(
            other,
            confidence_level=confidence_level,
        ):
            since_first_disparity \
                = self.CalcDefinitiveMinRanksSinceFirstRetainedDisparityWith(
                    other,
                )

            lb_exclusive = opyt.or_value(since_first_disparity, -1)
            lb_inclusive = lb_exclusive + 1

            since_last_commonality \
                = self.CalcRanksSinceLastRetainedCommonalityWith(
                    other,
                    confidence_level=confidence_level,
                )
            assert since_last_commonality is not None
            ub_inclusive = since_last_commonality
            ub_exclusive = ub_inclusive + 1

            return (lb_inclusive, ub_exclusive)
        else:
            return None

    def CalcRanksSinceMrcaUncertaintyWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float=0.95,
    ) -> typing.Optional[int]:
        """Calculate uncertainty of estimate for the number of depositions
        elapsed along this column's line of descent since the most common recent
        ancestor with other.

        Returns 0 if no common ancestor between self and other can be resolved
        with sufficient confidence. If insufficient common ranks between self
        and other are available to resolve any common ancestor, returns None.

        See Also
        --------
        CalcRanksSinceMrcaBoundsWith :
            Calculates bound whose uncertainty this method reports. See the
            corresponding docstring for explanation of parameters.
        """

        assert 0.0 <= confidence_level <= 1.0

        if self.CalcRankOfEarliestDetectableMrcaWith(
            other,
            confidence_level=confidence_level,
        ) is None:
            return None

        bounds = self.CalcRanksSinceMrcaBoundsWith(
            other,
            confidence_level=confidence_level,
        )
        return 0 if bounds is None else abs(operator.sub(*bounds)) - 1

    def CalcRanksSinceEarliestDetectableMrcaWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float=0.95,
    ) -> typing.Optional[int]:
        """How many depositions have elapsed since the earliest possible rank a
        MRCA between self and other could be reliably detected at?

        Even if a true MRCA of self and other exists, if it occured earlier
        than the rank calculated here it could not be reliably detected with
        sufficient confidence after accounting for the possibility of spurious
        differentia collisions. (Although subsequent spurious differentia
        collisions after the true MRCA of self and other could lead to MRCA
        detection at such a rank.)

        Returns None if insufficient common ranks exist between self and other
        to ever conclude at the given confidence level the existance of any
        common ancestry between self and other (even if all strata at common
        ranks had equivalent differentiae).
        """

        return opyt.apply_if(
            self.CalcRankOfEarliestDetectableMrcaWith(
                other,
                confidence_level=confidence_level,
            ),
            lambda rank: self.GetNumStrataDeposited() - 1 - rank,
        )

    def GetLastCommonStratumWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float=0.95,
    ) -> typing.Optional[HereditaryStratum]:
        """Get the most recent stratum in common between self and other, if any.

        Common strata share identical rank and differentia. Returns None if no
        common strata exist between the two columns. Allows probability equal
        to 1 - confidence_level that the last true common stratum is before the
        stratum returned (i.e., strata were erroneously detected as common due
        to spurious differentia collisions).

        See Also
        --------
        CalcRankOfLastRetainedCommonalityWith :
            Selects the stratum returned. See the corresponding docstring for
            explanation of parameters.
        """

        rank = self.CalcRankOfLastRetainedCommonalityWith(
            other,
            confidence_level=confidence_level,
        )
        if rank is not None:
            index = ip.popsingleton(
                index
                for index in range(self.GetNumStrataRetained())
                if rank == self.GetRankAtColumnIndex(index)
            )
            return self.GetStratumAtColumnIndex(index)
        else: return None

    def DefinitivelySharesNoCommonAncestorWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
    ) -> bool:
        """Does self definitively share no common ancestor with other?

        Note that stratum rention policies are strictly required to permanently
        retain the most ancient stratum.

        See Also
        --------
        HasAnyCommonAncestorWith:
            Can we conclude with confidence_level confidence that self and other
            share a common ancestor?
        """

        first_disparity \
            = self.CalcDefinitiveMaxRankOfFirstRetainedDisparityWith(other)
        return False if first_disparity is None else first_disparity == 0

    def HasAnyCommonAncestorWith(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
        confidence_level: float=0.95,
    ) -> typing.Optional[bool]:
        """Does self share any common ancestor with other?

        If insufficient common ranks between self and other are available to
        resolve any common ancestor, returns None.

        Note that stratum rention policies are strictly required to permanently
        retain the most ancient stratum.

        Parameters
        ----------
        confidence_level : float, optional
            The probability that we will correctly conclude no common ancestor
            is shared with other if, indeed, no common ancestor is actually
            shared. Default 0.95.

        See Also
        --------
        DefinitivelySharesNoCommonAncestorWith :
            Can we definitively conclude that self and other share no common
            ancestor?
        """

        if self.CalcRankOfEarliestDetectableMrcaWith(
            other,
            confidence_level=confidence_level,
        ) is None:
            return None

        first_disparity = self.CalcRankOfFirstRetainedDisparityWith(
            other,
            confidence_level=confidence_level,
        )
        return True if first_disparity is None else first_disparity > 0

    def Clone(
            self: 'HereditaryStratigraphicColumn',
    ) -> 'HereditaryStratigraphicColumn':
        """Create a copy of the store with identical data that may be freely
        altered without affecting data within this store."""

        # shallow copy
        result = copy(self)
        # do semi-shallow duplication on select elements
        result._stratum_ordered_store = self._stratum_ordered_store.Clone()
        return result

    def CloneDescendant(
        self: 'HereditaryStratigraphicColumn',
        stratum_annotation: typing.Optional[typing.Any]=None,
    ) -> 'HereditaryStratigraphicColumn':
        """Return a cloned bundle that has had an additional stratum deposited.

        Does not alter self.

        Parameters
        ----------
        stratum_annotation: any, optional
            Optional object to store as an annotation. Allows arbitrary user-
            provided to be associated with this stratum deposition in the
            line of descent.
        """

        res = self.Clone()
        res.DepositStratum(annotation=stratum_annotation)
        return res

    def DiffRetainedRanks(
        self: 'HereditaryStratigraphicColumn',
        other: 'HereditaryStratigraphicColumn',
    ) -> typing.Tuple[typing.Set[int], typing.Set[int]]:
        """Return the set of ranks retained by self but not other and vice
        versa as a tuple."""

        self_ranks = set( self.GetRetainedRanks() )
        other_ranks = set( other.GetRetainedRanks() )

        return (
            self_ranks - other_ranks,
            other_ranks - self_ranks,
        )



from copy import deepcopy
import random
import unittest

from hstrat import hstrat


class TestStratumRetentionPredicateDepthProportionalResolution(
    unittest.TestCase,
):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_equality(self):
        assert (
            hstrat.StratumRetentionPredicateDepthProportionalResolution()
            == hstrat.StratumRetentionPredicateDepthProportionalResolution()
        )

        original1 = hstrat.StratumRetentionPredicateDepthProportionalResolution(
            guaranteed_depth_proportional_resolution=10,
        )
        original2 = hstrat.StratumRetentionPredicateDepthProportionalResolution(
            guaranteed_depth_proportional_resolution=42,
        )
        copy1 = deepcopy(original1)
        assert original1 == copy1
        assert original1 != original2
        assert copy1 != original2

    def _do_test_space_complexity(self, guaranteed_depth_proportional_resolution):
        predicate = hstrat.StratumRetentionPredicateDepthProportionalResolution(
            guaranteed_depth_proportional_resolution
                =guaranteed_depth_proportional_resolution,
        )
        column = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_predicate=predicate,
        )

        for generation in range(10000):
            assert (
                column.GetNumStrataRetained()
                <= predicate.CalcNumStrataRetainedUpperBound(
                    num_strata_deposited=column.GetNumStrataDeposited(),
                )
            )
            column.DepositStratum()

    def test_space_complexity(self):
        for guaranteed_depth_proportional_resolution in [
            1,
            2,
            10,
            42,
            97,
        ]:
            self._do_test_space_complexity(
                guaranteed_depth_proportional_resolution,
            )

    def _do_test_resolution(
        self,
        guaranteed_depth_proportional_resolution,
        synchronous,
    ):
        predicate = hstrat.StratumRetentionPredicateDepthProportionalResolution(
            guaranteed_depth_proportional_resolution
                =guaranteed_depth_proportional_resolution,
        )
        column = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_predicate=predicate,
        )

        population = [
            column.Clone()
            for __ in range(25)
        ]

        for generation in range(500):

            # subsample consecutive pairs in population
            for f, s in  zip(population, population[1:]):
                target_resolu = predicate.CalcMrcaUncertaintyUpperBound(
                    first_num_strata_deposited=f.GetNumStrataDeposited(),
                    second_num_strata_deposited=s.GetNumStrataDeposited(),
                )
                assert f.CalcRankOfMrcaUncertaintyWith(s) <= target_resolu
                assert f.CalcRanksSinceMrcaUncertaintyWith(s) <= target_resolu

            random.shuffle(population)
            for target in range(5):
                population[target] = population[-1].Clone()
            for individual in population:
                if synchronous or random.choice([True, False]):
                    individual.DepositStratum()

    def test_resolution(self):
        for min_intervals_divide_into in [
            1,
            2,
            10,
            17,
        ]:
            for synchronous in True, False:
                self._do_test_resolution(min_intervals_divide_into, synchronous)

    def _do_test_CalcRankAtColumnIndex(
        self,
        guaranteed_depth_proportional_resolution,
    ):
        predicate = hstrat.StratumRetentionPredicateDepthProportionalResolution(
            guaranteed_depth_proportional_resolution
                =guaranteed_depth_proportional_resolution,
        )
        column = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_predicate=predicate,
            initial_stratum_annotation=0,
        )

        for generation in range(1,5001):
            for index in range(column.GetNumStrataRetained()):
                assert (
                    column.GetStratumAtColumnIndex(index).GetAnnotation()
                    == column.GetRankAtColumnIndex(index)
                )
            column.DepositStratum(annotation=generation)

    def test_CalcRankAtColumnIndex(self):
        for min_intervals_divide_into in [
            1,
            2,
            10,
            17,
        ]:
            self._do_test_CalcRankAtColumnIndex(min_intervals_divide_into)


if __name__ == '__main__':
    unittest.main()



from copy import deepcopy
from iterify import cyclify
import itertools as it
import random
import unittest

random.seed(1)

from hstrat import hstrat


class TestStratumRetentionPredicateRecencyProportionalResolution(
    unittest.TestCase,
):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_equality(self):
        assert (
            hstrat.StratumRetentionPredicateRecencyProportionalResolution()
            == hstrat.StratumRetentionPredicateRecencyProportionalResolution()
        )

        original1 \
            = hstrat.StratumRetentionPredicateRecencyProportionalResolution(
                guaranteed_mrca_recency_proportional_resolution=2,
        )
        original2 \
            = hstrat.StratumRetentionPredicateRecencyProportionalResolution(
                guaranteed_mrca_recency_proportional_resolution=3,
        )
        copy1 = deepcopy(original1)
        assert original1 == copy1
        assert original1 != original2
        assert copy1 != original2

    def _do_test_space_complexity(
        self,
        guaranteed_mrca_recency_proportional_resolution,
    ):
        predicate = \
            hstrat.StratumRetentionPredicateRecencyProportionalResolution(
                guaranteed_mrca_recency_proportional_resolution
                    =guaranteed_mrca_recency_proportional_resolution,
        )
        column = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_predicate=predicate,
            stratum_ordered_store_factory
                =hstrat.HereditaryStratumOrderedStoreList,
        )

        for generation in range(1, 1000):
            target_space = predicate.CalcNumStrataRetainedExact(generation)
            assert target_space == column.GetNumStrataRetained()
            column.DepositStratum()

    def test_space_complexity(self):
        for guaranteed_mrca_recency_proportional_resolution in [
            0,
            1,
            2,
            10,
            42,
        ]:
            self._do_test_space_complexity(
                guaranteed_mrca_recency_proportional_resolution,
            )


    def _do_test_resolution(
        self,
        guaranteed_mrca_recency_proportional_resolution,
        synchronous,
    ):
        test_predicate = \
            hstrat.StratumRetentionPredicateRecencyProportionalResolution(
                guaranteed_mrca_recency_proportional_resolution
                    =guaranteed_mrca_recency_proportional_resolution,
        )
        column_test = hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store_factory
                =hstrat.HereditaryStratumOrderedStoreList,
            stratum_retention_predicate=test_predicate,
        )
        column_control = hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store_factory
                =hstrat.HereditaryStratumOrderedStoreList,
            stratum_retention_condemner
                =hstrat.StratumRetentionCondemnerPerfectResolution(),
        )

        column_bundle = hstrat.HereditaryStratigraphicColumnBundle({
            'test' : column_test,
            'control' : column_control,
        })

        population = [
            column_bundle.Clone()
            for __ in range(25)
        ]
        forked_population = [
            column_bundle.Clone()
            for __ in range(5)
        ]
        ancestor = [ column_bundle.Clone() ]

        for generation in range(500):
            # subsample consecutive pairs in population
            for f, s in it.chain(
                zip(population, population[1:]),
                zip(forked_population, population),
                zip(population, forked_population),
                zip(ancestor, population),
                zip(population, ancestor),
            ):

                actual_rank_of_mrca \
                    = f['control'].CalcRankOfLastRetainedCommonalityWith(
                        s['control']
                    )
                assert (
                    f['control'].CalcRankOfMrcaUncertaintyWith(s['control'])
                    == 0
                )

                target_resolu = test_predicate.CalcMrcaUncertaintyUpperBound(
                    actual_rank_of_mrca=actual_rank_of_mrca,
                    first_num_strata_deposited=f.GetNumStrataDeposited(),
                    second_num_strata_deposited=s.GetNumStrataDeposited(),
                )

                assert (
                    f['test'].CalcRankOfMrcaUncertaintyWith(s['test'])
                    <= target_resolu
                )
                assert (
                    s['test'].CalcRankOfMrcaUncertaintyWith(f['test'])
                    <= target_resolu
                )
                assert (
                    f['test'].CalcRanksSinceMrcaUncertaintyWith(s['test'])
                    <= target_resolu
                )
                assert (
                    s['test'].CalcRanksSinceMrcaUncertaintyWith(f['test'])
                    <= target_resolu
                )

            random.shuffle(population)
            random.shuffle(forked_population)
            # reproduction
            for target in range(5):
                population[target] = population[-1].Clone()
            # advance generations
            for individual in it.chain(
                iter(population),
                iter(forked_population),
            ):
                if synchronous or random.choice([True, False]):
                    individual.DepositStratum()

    def test_resolution(self):
        for guaranteed_mrca_recency_proportional_resolution in [
            0,
            1,
            2,
            3,
            17,
            100,
        ]:
            for synchronous in True, False:
                self._do_test_resolution(
                    guaranteed_mrca_recency_proportional_resolution,
                    synchronous,
                )

    def _do_test_resolution_regression_case(
        self,
        actual_rank_of_mrca,
        guaranteed_mrca_recency_proportional_resolution,
        num_strata_deposited,
    ):
        test_predicate = \
            hstrat.StratumRetentionPredicateRecencyProportionalResolution(
                guaranteed_mrca_recency_proportional_resolution
                    =guaranteed_mrca_recency_proportional_resolution,
        )
        first = hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store_factory
                =hstrat.HereditaryStratumOrderedStoreDict,
            stratum_retention_predicate=test_predicate,
        )

        for __ in range(actual_rank_of_mrca): first.DepositStratum()

        second = first.Clone()

        first_num_strata, second_num_strata = num_strata_deposited
        for __ in range(first_num_strata - actual_rank_of_mrca):
            first.DepositStratum()
        for __ in range(second_num_strata - actual_rank_of_mrca):
            second.DepositStratum()

        target_resolution = test_predicate.CalcMrcaUncertaintyUpperBound(
            actual_rank_of_mrca=actual_rank_of_mrca,
            first_num_strata_deposited=first.GetNumStrataDeposited(),
            second_num_strata_deposited=second.GetNumStrataDeposited(),
        )

        assert (
            second.CalcRankOfMrcaUncertaintyWith(first)
            <= target_resolution
        )
        assert (
            first.CalcRanksSinceMrcaUncertaintyWith(second)
            <= target_resolution
        )
        assert (
            second.CalcRanksSinceMrcaUncertaintyWith(first)
            <= target_resolution
        )

    def test_resolution_regression(self):
        for case in [
            {
                'actual_rank_of_mrca' : 111,
                'guaranteed_mrca_recency_proportional_resolution' : 17,
                'num_strata_deposited' : (148, 152),
            },
            {
                'actual_rank_of_mrca' : 101,
                'guaranteed_mrca_recency_proportional_resolution' : 17,
                'num_strata_deposited' : (144, 144),
            },
            {
                'actual_rank_of_mrca' : 103,
                'guaranteed_mrca_recency_proportional_resolution' : 17,
                'num_strata_deposited' : (144, 144),
            },
        ]:
            self._do_test_resolution_regression_case(**case)

    def _do_test_deep_resolution(
        self,
        guaranteed_mrca_recency_proportional_resolution,
    ):
        test_predicate = \
            hstrat.StratumRetentionPredicateRecencyProportionalResolution(
                guaranteed_mrca_recency_proportional_resolution
                    =guaranteed_mrca_recency_proportional_resolution,
        )
        column_test = hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store_factory
                =hstrat.HereditaryStratumOrderedStoreDict,
            stratum_retention_predicate=test_predicate,
        )
        column_control = hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store_factory
                =hstrat.HereditaryStratumOrderedStoreList,
            stratum_retention_condemner
                =hstrat.StratumRetentionCondemnerPerfectResolution(),
        )

        individual = hstrat.HereditaryStratigraphicColumnBundle({
            'test' : column_test,
            'control' : column_control,
        })

        snapshots = [individual]
        for snapshot in range(100):
                for fastforward in range(100):
                        individual.DepositStratum()
                snapshots.append(individual.Clone())


        for f, s in it.chain(
            zip(cyclify(individual), snapshots),
            zip(snapshots, cyclify(individual)),
        ):
            actual_rank_of_mrca = (
                f['control'].CalcRankOfLastRetainedCommonalityWith(s['control'])
            )

            target_resolu = test_predicate.CalcMrcaUncertaintyUpperBound(
                actual_rank_of_mrca=actual_rank_of_mrca,
                first_num_strata_deposited=f.GetNumStrataDeposited(),
                second_num_strata_deposited=s.GetNumStrataDeposited(),
            )

            assert (
                f['test'].CalcRankOfMrcaUncertaintyWith(s['test'])
                <= target_resolu
            )
            assert (
                s['test'].CalcRankOfMrcaUncertaintyWith(f['test'])
                <= target_resolu
            )
            assert (
                f['test'].CalcRanksSinceMrcaUncertaintyWith(s['test'])
                <= target_resolu
            )
            assert (
                s['test'].CalcRanksSinceMrcaUncertaintyWith(f['test'])
                <= target_resolu
            )

    def test_deep_resolution(self):
        for guaranteed_mrca_recency_proportional_resolution in [
            0,
            1,
            2,
            3,
            17,
            100,
        ]:
            self._do_test_deep_resolution(
                guaranteed_mrca_recency_proportional_resolution,
            )

    def _do_test_deep_space_complexity(
        self,
        guaranteed_mrca_recency_proportional_resolution,
    ):
        test_predicate = \
            hstrat.StratumRetentionPredicateRecencyProportionalResolution(
                guaranteed_mrca_recency_proportional_resolution
                    =guaranteed_mrca_recency_proportional_resolution,
            )
        individual = hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store_factory
                =hstrat.HereditaryStratumOrderedStoreDict,
            stratum_retention_predicate=test_predicate,
        )

        snapshots = [individual]
        for snapshot in range(100):
                for fastforward in range(100):
                        individual.DepositStratum()
                snapshots.append(individual.Clone())


        for snapshot in snapshots:
            target_space = test_predicate.CalcNumStrataRetainedExact(
                num_strata_deposited=snapshot.GetNumStrataDeposited(),
            )
            # implementation provides an exact count of num retained
            assert snapshot.GetNumStrataRetained() == target_space

    def test_deep_space_complexity(self):
        for guaranteed_mrca_recency_proportional_resolution in [
            1,
            2,
            3,
            17,
            100,
        ]:
            self._do_test_deep_space_complexity(
                guaranteed_mrca_recency_proportional_resolution,
            )


if __name__ == '__main__':
    unittest.main()



from copy import copy
import math
import operator
import typing

from .HereditaryStratigraphicColumn import HereditaryStratigraphicColumn


class HereditaryStratigraphicColumnBundle:
    """Packages multiple HereditaryStratigraphicColumn instances together in
    order to conveniently advance them in sync along a line of descent with a
    similar interface to an individual HereditaryStratigraphicColumn.
    """

    _columns: typing.Dict[str, HereditaryStratigraphicColumn]

    def __init__(
        self: 'HereditaryStratigraphicColumnBundle',
        columns: typing.Dict[str, HereditaryStratigraphicColumn],
    ):
        """Construct bundle.

        Parameters
        ----------
        columns : dict
            HereditaryStratigraphicColumn objects to bundle together, each
            associated with a unique names as its key.
        """
        assert len(columns), \
            "Must provide at least one column to " \
            "HereditaryStratigraphicColumnBundle."
        assert len({c.GetNumStrataDeposited() for c in columns.values()}) == 1,\
            "All columns provided ot HereditaryStratigraphicColumnBundle " \
            "must have same number strata deposited. "
        self._columns = columns

    def __iter__(
        self: 'HereditaryStratigraphicColumnBundle',
    ) -> typing.Iterator[str]:
        """Iterable interface."""

        yield from self._columns

    def __getitem__(
        self: 'HereditaryStratigraphicColumnBundle',
        key: str,
    ) -> HereditaryStratigraphicColumn:
        """Brackets operator; access constituent column by name."""

        return self._columns[key]

    def __eq__(
        self: 'HereditaryStratigraphicColumnBundle',
        other: 'HereditaryStratigraphicColumnBundle',
    ) -> bool:
        """Compare for value-wise equality."""

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def DepositStratum(
        self: 'HereditaryStratigraphicColumnBundle',
        annotation: typing.Optional[typing.Any]=None,
    ) -> None:
        """Elapse a generation on all constituent columns.

        Parameters
        ----------
        annotation: any, optional
            Optional object to store as an annotation. Allows arbitrary user-
            provided to be associated with this stratum's generation in its
            line of descent.
        """
        for column in self._columns.values():
            column.DepositStratum(
                annotation=annotation,
            )

    def GetNumStrataDeposited(
        self: 'HereditaryStratigraphicColumnBundle',
    ) -> int:
        """How many strata have been deposited on constituent columns?

        Should be identical across constituent columns."""
        return next(iter(self._columns.values())).GetNumStrataDeposited()

    def Clone(
            self: 'HereditaryStratigraphicColumnBundle',
    ) -> 'HereditaryStratigraphicColumnBundle':
        """Create a copy of the bundle with identical data that may be freely
        altered without affecting data within this bundle."""

        # shallow copy
        result = copy(self)
        # do semi-shallow clone on select elements
        # see https://stackoverflow.com/a/5861653 for performance consierations
        result._columns = {
            k : v.Clone()
            for k, v in self._columns.items()
        }
        return result

    def CloneDescendant(
        self: 'HereditaryStratigraphicColumnBundle',
        stratum_annotation: typing.Optional[typing.Any]=None,
    ) -> 'HereditaryStratigraphicColumnBundle':
        """Return a cloned bundle that has had an additional stratum deposited.

        Does not alter self.

        Parameters
        ----------
        stratum_annotation: any, optional
            Optional object to store as an annotation. Allows arbitrary user-
            provided to be associated with this stratum deposition in the
            line of descent.
        """

        res = self.Clone()
        res.DepositStratum(annotation=stratum_annotation)
        return res

    def __getattr__(self, attr):
        """Forward all unknown method calls and property accesses to underlying
        HereditaryStratigraphicColumns, returning a dict of results for each
        column key stored.

        Note that __getattr__ is only called after other attribute lookup (i.e.,
        explicitly provided methods and properties) has failed.
        """
        # Adapted from
        # https://rosettacode.org/wiki/Respond_to_an_unknown_method_call#Python

        # raise AttributeError for dunder methods so that callers expecting
        # them unimplemented can catch and run their fallbacks
        if '__' in attr:
            raise AttributeError

        def arg_debundler(args, column_name):
            """If any args are column bundles, extract the focal column."""

            return [
                arg
                if not isinstance(arg, self.__class__)
                else arg[column_name]
                for arg in args
            ]

        def kwarg_debundler(kwargs, column_name):
            """If any kwarg vals are column bundles, extract the focal
            column."""

            return {
                k : v \
                    if not isinstance(v, self.__class__)
                    else v[column_name]
                for k, v in kwargs.items()
            }

        if any(
            callable(getattr(column, attr))
            for column in self._columns.values()
        ):
            # method forwarding
            assert all(
                callable(getattr(column, attr))
                for column in self._columns.values()
            )

            def forwarded(*args, **kwargs):
                """Apply method to each column independently, extracting the
                corresponding column from any column bundles passed as
                arguments.

                Returns a dict mapping each column name to its result.
                """

                return {
                    column_name : getattr(column, attr)(
                        *arg_debundler(args, column_name),
                        **kwarg_debundler(kwargs, column_name),
                    )
                    for column_name, column in self._columns.items()
                }

            return forwarded

        else:
            # property forwarding
            return {
                column_name : getattr(column, attr)
                for column_name, column in self._columns.items()
            }



from copy import deepcopy
import unittest

from hstrat import hstrat


class TestStratumRetentionPredicateStochastic(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_equality(self):
        assert (
            hstrat.StratumRetentionPredicateStochastic()
            == hstrat.StratumRetentionPredicateStochastic()
        )

        original = hstrat.StratumRetentionPredicateStochastic()
        copy = deepcopy(original)
        assert original == copy


if __name__ == '__main__':
    unittest.main()



import typing

from ..HereditaryStratum import HereditaryStratum
from ..stratum_retention_predicates \
    import StratumRetentionPredicateRecencyProportionalResolution


class StratumRetentionCondemnerRecencyProportionalResolution(
    # inherit CalcNumStrataRetainedUpperBound, etc.
    StratumRetentionPredicateRecencyProportionalResolution,
):
    """Functor to implement the MRCA-recency-proportional resolution stratum retention policy, for use with HereditaryStratigraphicColumn.

    This functor enacts the MRCA-recency-proportional resolution policy by
    specifying the set of strata ranks that should be purged from a hereditary
    stratigraphic column when the nth stratum is deposited.

    The MRCA-recency-proportional resolution policy ensures estimates of MRCA
    rank will have uncertainty bounds less than or equal to a user-specified
    proportion of the actual number of generations elapsed since the MRCA and
    the deepest of the compared columns. MRCA rank estimate uncertainty scales
    in the worst case scales as O(n) with respect to the greater number of
    strata deposited on either column. However, with respect to estimating the rank of the MRCA when lineages diverged any fixed number of generations ago,
    uncertainty scales as O(1).

    Under the MRCA-recency-proportional resolution policy, the number of strata
    retained (i.e., space complexity) scales as O(log(n)) with respect to the
    number of strata deposited.

    See Also
    --------
    StratumRetentionPredicateRecencyProportionalResolution:
        For definitions of methods inherited by this class that describe
        guaranteed properties of the MRCA-recency proportional resolution
        stratum retention policy.
    """

    def __init__(
        self: 'StratumRetentionCondemnerRecencyProportionalResolution',
        guaranteed_mrca_recency_proportional_resolution: int=10,
    ):
        """Construct the functor.

        Parameters
        ----------
        guaranteed_mrca_recency_proportional_resolution : int, optional
            The desired minimum number of intervals between the MRCA and the
            deeper compared column to be able to be distinguished between. The
            uncertainty of MRCA rank estimates provided under the MRCA-recency-
            proportional resolution policy will scale as the actual
            phylogenetic depth of the MRCA divided by
            guaranteed_mrca_recency_proportional_resolution.
        """

        super(
            StratumRetentionCondemnerRecencyProportionalResolution,
            self,
        ).__init__(
            guaranteed_mrca_recency_proportional_resolution
                =guaranteed_mrca_recency_proportional_resolution,
        )

    def _num_to_condemn(
        self: 'StratumRetentionCondemnerRecencyProportionalResolution',
        num_stratum_depositions_completed: int,
    ) -> int:
        """How many strata should be eliminated after
        num_stratum_depositions_completed have been deposited and the
        num_stratum_depositions_completed + 1'th deposition is in progress?

        Used to implement functor's __call__ method specifying which ranks
        should be purged during this stratum deposition. This expression for
        exact number deposited was extrapolated from
            * resolution = 0, <https://oeis.org/A001511>
            * resolution = 1, <https://oeis.org/A091090>
        and is unit tested extensively.
        """

        resolution = self._guaranteed_mrca_recency_proportional_resolution
        # _guaranteed_mrca_recency_proportional_resolution is from super class

        if num_stratum_depositions_completed % 2 == 1:
            return 0
        elif num_stratum_depositions_completed < 2 * (resolution + 1):
            return 0
        else:
            return \
                1 + self._num_to_condemn(num_stratum_depositions_completed // 2)

    def __call__(
        self: 'StratumRetentionCondemnerRecencyProportionalResolution',
        num_stratum_depositions_completed: int,
        retained_ranks: typing.Optional[typing.Iterable[int]]=None,
    ) -> typing.Iterator[int]:
        """Decide which strata within the stratagraphic column should be purged.

        Every time a new stratum is deposited, this method is called to
        determine which strata should be purged. All strata at ranks yielded
        from this functor are immediately purged from the column, meaning that
        for a stratum to persist it must not be yielded by this functor each
        and every time a new stratum is deposited.

        Parameters
        ----------
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.
        retained_ranks : iterator over int, optional
            An iterator over ranks of strata currently retained within the
            hereditary stratigraphic column. Not used in this functor.

        Returns
        -------
        iterator over int
                The ranks of strata that should be purged from the hereditary
            stratigraphic column at this deposition step.

        See Also
        --------
        StratumRetentionPredicateRecencyProportionalResolution:
            For details on the rationale, implementation, and guarantees of the
            recency-proportional resolution stratum retention policy.
        """

        num_to_condemn = self._num_to_condemn(num_stratum_depositions_completed)
        resolution = self._guaranteed_mrca_recency_proportional_resolution
        # _guaranteed_mrca_recency_proportional_resolution is from super class

        for i in range(num_to_condemn):
            factor = 2 * resolution + 1
            num_ranks_back = factor * (2 ** i)
            yield num_stratum_depositions_completed - num_ranks_back



from .HereditaryStratigraphicColumn import HereditaryStratigraphicColumn
from .HereditaryStratigraphicColumnBundle \
    import HereditaryStratigraphicColumnBundle
from .HereditaryStratum import HereditaryStratum
from .HereditaryStratumOrderedStoreDict import HereditaryStratumOrderedStoreDict
from .HereditaryStratumOrderedStoreList import HereditaryStratumOrderedStoreList
from .HereditaryStratumOrderedStoreTree import HereditaryStratumOrderedStoreTree
from .stratum_retention_condemners import *
from .stratum_retention_predicates import *

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    'HereditaryStratigraphColumn',
    'HereditaryStratigraphicColumnBundle',
    'HereditaryStratum',
    'HereditaryStratumOrderedStoreDict',
    'HereditaryStratumOrderedStoreList',
    'HereditaryStratumOrderedStoreTree',
]



from .AnyTreeAscendingIter import AnyTreeAscendingIter
from .attach_attrs import attach_attrs

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    'AnyTreeAscendingIter',
    'attach_attrs',
]



from copy import deepcopy
import unittest

from hstrat import hstrat


class TestStratumRetentionCondemnerPerfectResolution(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_equality(self):
        assert (
            hstrat.StratumRetentionCondemnerPerfectResolution()
            == hstrat.StratumRetentionCondemnerPerfectResolution()
        )

        original = hstrat.StratumRetentionCondemnerPerfectResolution()
        copy = deepcopy(original)
        assert original == copy

    def test_condemnation(self):
        maximal_retention_condemner \
            = hstrat.StratumRetentionCondemnerPerfectResolution()
        for i in range(100):
            assert list(maximal_retention_condemner(
                retained_ranks=range(i),
                num_stratum_depositions_completed=i,
            )) == []

    def test_retention(self):
        control_column = hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store_factory
                =hstrat.HereditaryStratumOrderedStoreDict,
            stratum_retention_predicate
                =hstrat.StratumRetentionPredicatePerfectResolution(),
        )
        test_column = hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store_factory
                =hstrat.HereditaryStratumOrderedStoreDict,
            stratum_retention_condemner
                =hstrat.StratumRetentionCondemnerPerfectResolution(),
        )

        for i in range(1000):
            control_column.DepositStratum()
            test_column.DepositStratum()
            d1, d2 = control_column.DiffRetainedRanks(test_column)
            assert d1 == set() and d2 == set()

    def test_CalcRankAtColumnIndex(self):
        test_condemner = hstrat.StratumRetentionCondemnerPerfectResolution()
        test_column = hstrat.HereditaryStratigraphicColumn(
            always_store_rank_in_stratum=True,
            stratum_ordered_store_factory
                =hstrat.HereditaryStratumOrderedStoreDict,
            stratum_retention_condemner=test_condemner,
        )

        for i in range(1000):
            test_column.DepositStratum()
            actual_ranks = { *test_column.GetRetainedRanks() }
            calcualted_ranks = {
                test_condemner.CalcRankAtColumnIndex(
                    index=i,
                    num_strata_deposited=test_column.GetNumStrataDeposited()
                )
                for i in range(test_column.GetNumStrataRetained())
            }
            assert actual_ranks == calcualted_ranks
            # in-progress deposition case
            assert test_condemner.CalcRankAtColumnIndex(
                index=test_column.GetNumStrataRetained(),
                num_strata_deposited=test_column.GetNumStrataDeposited(),
            ) == test_column.GetNumStrataDeposited()

    def test_CalcNumStrataRetainedExact(self):
        test_condemner \
            = hstrat.StratumRetentionCondemnerPerfectResolution()
        test_column = hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store_factory
                =hstrat.HereditaryStratumOrderedStoreDict,
            stratum_retention_condemner=test_condemner,
        )

        assert 0 == test_condemner.CalcNumStrataRetainedExact(
            num_strata_deposited=0,
        )
        for i in range(10000):
            calculated_num_retained = test_condemner.CalcNumStrataRetainedExact(
                num_strata_deposited=test_column.GetNumStrataDeposited(),
            )
            observed_num_retained = test_column.GetNumStrataRetained()
            assert calculated_num_retained == observed_num_retained
            test_column.DepositStratum()

    def test_CalcNumStrataRetainedUpperBound(self):
        test_condemner \
            = hstrat.StratumRetentionCondemnerPerfectResolution()
        test_column = hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store_factory
                =hstrat.HereditaryStratumOrderedStoreDict,
            stratum_retention_condemner=test_condemner,
        )

        assert 0 <= test_condemner.CalcNumStrataRetainedUpperBound(
            num_strata_deposited=0,
        )
        for i in range(10000):
            calculated_num_retained_bound \
                = test_condemner.CalcNumStrataRetainedUpperBound(
                    num_strata_deposited=test_column.GetNumStrataDeposited(),
            )
            observed_num_retained = test_column.GetNumStrataRetained()
            assert calculated_num_retained_bound >= observed_num_retained
            test_column.DepositStratum()


if __name__ == '__main__':
    unittest.main()



import math
import typing


class StratumRetentionPredicateDepthProportionalResolution:
    """Functor to implement the depth-proportional resolution stratum retention
    policy, for use with HereditaryStratigraphicColumn.

    This functor enacts the depth-proportional resolution policy by specifying
    whether a stratum with deposition rank r should be retained within the
    hereditary stratigraphic column after n strata have been deposited.

    The depth-proportional resolution policy ensures estimates of MRCA rank will
    have uncertainty bounds less than or equal to a user-specified
    proportion of the largest number of strata deposited on either column.
    Thus, MRCA rank estimate uncertainty scales as O(n) with respect to the
    greater number of strata deposited on either column.

    Under the depth-proportional resolution policy, the number of strata
    retained (i.e., space complexity) scales as O(1) with respect to the number
    of strata deposited.

    See Also
    --------
    StratumRetentionCondemnerDepthProportionalResolution:
        For a potentially more computationally efficient specificiation of the
        depth-proportional resolution policy that directly generates the ranks
        of strata that should be purged during the nth stratum deposition.
    """

    _guaranteed_depth_proportional_resolution: int

    def __init__(
        self: 'StratumRetentionPredicateDepthProportionalResolution',
        guaranteed_depth_proportional_resolution: int=10
    ):
        """Construct the functor.

        Parameters
        ----------
        guaranteed_depth_proportional_resolution : int, optional
            The desired minimum number of intervals for the rank of the MRCA to
            be able to be distinguished between. The uncertainty of MRCA
            rank estimates provided under the depth-proportional resolution
            policy will scale as total number of strata deposited divided by
            guaranteed_depth_proportional_resolution.
        """

        assert guaranteed_depth_proportional_resolution > 0
        self._guaranteed_depth_proportional_resolution = (
            guaranteed_depth_proportional_resolution
        )

    def __eq__(
        self: 'StratumRetentionPredicateDepthProportionalResolution',
        other: 'StratumRetentionPredicateDepthProportionalResolution',
    ) -> bool:
        """Compare for value-wise equality."""

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def _calc_provided_uncertainty(
        self: 'StratumRetentionPredicateDepthProportionalResolution',
        num_stratum_depositions_completed: int,
    ) -> int:
        """After n strata have been deposited, how many ranks are spaced
        between retained strata?"""

        guaranteed_resolution = self._guaranteed_depth_proportional_resolution
        max_uncertainty \
            = num_stratum_depositions_completed // guaranteed_resolution

        # round down to lower or equal power of 2
        provided_uncertainty_exp = (max_uncertainty // 2).bit_length()
        provided_uncertainty = 2 ** provided_uncertainty_exp
        return provided_uncertainty

    def __call__(
        self: 'StratumRetentionPredicateDepthProportionalResolution',
        stratum_rank: int,
        num_stratum_depositions_completed: int,
    ) -> bool:
        """Decide if a stratum within the stratagraphic column should be
        retained or purged.

        Every time a new stratum is deposited, this method is called on each
        stratum present in a HereditaryStratigraphicColumn to determine whether
        it should be retained. Strata that return False are immediately purged
        from the column, meaning that for a stratum to persist it must earn a
        True result from this method each and every time a new stratum is
        deposited.

        This functor's retention policy implementation guarantees that columns
        will retain appropriate strata so that for any two columns with m and n
        strata deposited, the rank of the most recent common ancestor can be
        determined with uncertainty of at most

            bound = floor(max(m, n) / guaranteed_depth_proportional_resolution)

        ranks. Achieving this limit on uncertainty requires retaining sufficient
        strata so that no more than bound ranks elapsed between any two strata.
        This policy accumulates retained strata at a fixed interval until twice
        as many as guaranteed_depth_proportional_resolution are at hand. Then,
        every other retained stratum is purged and the cycle repeats with a new
        twice-as-wide interval between retained strata.

        Suppose guaranteed_depth_proportional_resolution is 3.

               guaranteed   actual
        time   resolution   uncertainty   column
        --------------------------------------------------------
        1      0            0             |
        2      0            0             ||
        3      1            0             |||
        4      1            0             ||||
        5      1            0             |||||
        6      2            2             | | ||
        7      2            2             | | | |
        8      2            2             | | | ||
        9      3            2             | | | | |
        10     3            2             | | | | ||
        11     3            2             | | | | | |
        12     4            4             |   |   |  |
        13     4            4             |   |   |   |
        14     4            4             |   |   |   ||
        15     5            4             |   |   |   | |
        16     5            4             |   |   |   |  |
        17     5            4             |   |   |   |   |
        18     6            4             |   |   |   |   ||
        19     6            4             |   |   |   |   | |
        20     6            4             |   |   |   |   |  |
        21     7            4             |   |   |   |   |   |
        22     7            4             |   |   |   |   |   ||
        23     7            4             |   |   |   |   |   | |
        24     8            8             |       |       |      |
        25     8            8             |       |       |       |
        26     8            8             |       |       |       ||
        27     9            8             |       |       |       | |
        28     9            8             |       |       |       |  |
        29     9            8             |       |       |       |   |
        30     10           8             |       |       |       |    |

        Parameters
        ----------
        stratum_rank : int
            The number of strata that were deposited before the stratum under
            consideration for retention.
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.

        Returns
        -------
        bool
            True if the stratum should be retained, False otherwise.
        """

        guaranteed_resolution = self._guaranteed_depth_proportional_resolution

        # easy edge cases we must always retain
        if (
            # always retain newest stratum
            stratum_rank == num_stratum_depositions_completed
            # retain all strata until more than num_intervals are deposited
            or num_stratum_depositions_completed < guaranteed_resolution
        ): return True

        # +1 because of in-progress deposition
        provided_uncertainty = self._calc_provided_uncertainty(
            num_stratum_depositions_completed + 1,
        )
        return stratum_rank % provided_uncertainty == 0

    def CalcNumStrataRetainedExact(
        self: 'StratumRetentionPredicateDepthProportionalResolution',
        num_strata_deposited: int,
    ) -> int:
        """Exactly how many strata are retained after n deposted?"""

        if num_strata_deposited == 0: return 0

        provided_uncertainty = self._calc_provided_uncertainty(
            num_strata_deposited,
        )
        newest_stratum_rank = num_strata_deposited - 1
        # +1 for 0'th rank stratum
        num_strata_at_uncertainty_intervals \
            = newest_stratum_rank // provided_uncertainty + 1
        newest_stratum_distinct_from_uncertainty_intervals \
            = (newest_stratum_rank % provided_uncertainty != 0)
        return (
            num_strata_at_uncertainty_intervals
            + newest_stratum_distinct_from_uncertainty_intervals
        )

    def CalcNumStrataRetainedUpperBound(
        self: 'StratumRetentionPredicateDepthProportionalResolution',
        num_strata_deposited: typing.Optional[int]=None,
    ) -> int:
        """At most, how many strata are retained after n deposted? Inclusive."""

        return self._guaranteed_depth_proportional_resolution * 2 + 1

    def CalcMrcaUncertaintyUpperBound(
        self: 'StratumRetentionPredicateDepthProportionalResolution',
        *,
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
        actual_rank_of_mrca: typing.Optional[int]=None,
    ) -> int:
        """At most, how much uncertainty to estimate rank of MRCA? Inclusive."""

        return max(
            first_num_strata_deposited,
            second_num_strata_deposited,
        ) // self._guaranteed_depth_proportional_resolution


    def _CalcRankAtColumnIndexImpl(
        self: 'StratumRetentionPredicateDepthProportionalResolution',
        index: int,
        num_strata_deposited: int,
    ) -> int:
        """After n strata have been deposited, what will the rank of the
        stratum at column index k be?

        Assumes the no in-progress stratum depositions that haven't been
        reflected in num_strata_deposited.
        """

        provided_uncertainty = self._calc_provided_uncertainty(
            num_strata_deposited,
        )
        return min(
            index * provided_uncertainty,
            num_strata_deposited - 1
        )

    def CalcRankAtColumnIndex(
        self: 'StratumRetentionPredicateDepthProportionalResolution',
        index: int,
        num_strata_deposited: int,
    ) -> int:
        """After n strata have been deposited, what will the rank of the
        stratum at column index k be?

        Enables a HereditaryStratigraphicColumn using this predicate to
        optimize away storage of rank annotations on strata. Takes into the
        account the possiblity for in-progress stratum depositions that haven't
        been reflected in num_strata_deposited.
        """

        # 0th index is always rank 0
        if index == 0: return 0

        prev = self._CalcRankAtColumnIndexImpl(index - 1, num_strata_deposited)
        cur = self._CalcRankAtColumnIndexImpl(index, num_strata_deposited)
        if (cur == prev):
            # in cases where the same rank result is calculated for this index
            # and preceding index, the current index is an in-progress
            # deposition and must be calculated as the rank succeeding the
            # previous stratum's rank
            return prev + 1
        else:
            return cur



from copy import copy
from interval_search import binary_search
import typing

from .HereditaryStratum import HereditaryStratum


class HereditaryStratumOrderedStoreList:
    """Container for use with HereditaryStratigraphicColumn to store deposited
    strata, implemented as a list.

    Retained strata are stored from most ancient (index 0, front) to most
    recent (back). Cloned stores instantiate an independent list (although
    strata are not deepcopied themselves).

    Potentially useful in scenarios where moderate strata counts are retained,
    many strata are deposited without column cloning, deleted strata tend to
    be more recent (i.e., not more ancient and toward the front of the list),
    or many comparisons to estimate most recent common ancestor are made
    between stratigraphic columns.
    """

    # strata stored from most ancient (index 0, front) to most recent (back)
    _data: typing.List[HereditaryStratum]

    def __init__(self: 'HereditaryStratumOrderedStoreList'):
        """Initialize instance variables."""

        self._data = []

    def __eq__(
        self: 'HereditaryStratumOrderedStoreList',
        other: 'HereditaryStratumOrderedStoreList',
    ) -> bool:
        """Compare for value-wise equality."""

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def DepositStratum(
        self: 'HereditaryStratumOrderedStoreList',
        rank: int,
        stratum: 'HereditaryStratum',
    ) -> None:
        """Insert a new stratum into the store.

        Parameters
        ----------
        rank : int
            The position of the stratum being deposited within the sequence of strata deposited into the column. Precisely, the number of strata that have been deposited before stratum.
        stratum : HereditaryStratum
            The stratum to deposit.
        """

        self._data.append(stratum)

    def GetNumStrataRetained(self: 'HereditaryStratumOrderedStoreList') -> int:
        """How many strata are present in the store?

        May be fewer than the number of strata deposited if deletions have
        occured.
        """

        return len(self._data)

    def GetStratumAtColumnIndex(
        self: 'HereditaryStratumOrderedStoreList',
        index: int,
        # needed for other implementations
        get_rank_at_column_index: typing.Optional[typing.Callable]=None,
    ) -> HereditaryStratum:
        """Get the stratum positioned at index i among retained strata.

        Index order is from most ancient (index 0) to most recent.

        Parameters
        ----------
        ranks : iterator over int
            The ranks that to be deleted.
        get_column_index_of_rank : callable, optional
            Callable that returns the index position within retained strata of
            the stratum deposited at rank r.
        """

        return self._data[index]

    def GetRankAtColumnIndex(
        self: 'HereditaryStratumOrderedStoreList',
        index: int,
    ) -> int:
        """What is the deposition rank of the stratum positioned at index i
        among retained strata?

        Index order is from most ancient (index 0) to most recent.
        """

        res_rank = self.GetStratumAtColumnIndex(index).GetDepositionRank()
        assert res_rank is not None
        return res_rank

    def GetColumnIndexOfRank(
        self: 'HereditaryStratumOrderedStoreList',
        rank: int,
    ) -> typing.Optional[int]:
        """What is the index position within retained strata of the stratum
        deposited at rank r?

        Returns None if no stratum with rank r is present within the store.
        """

        if self.GetNumStrataRetained() == 0:
            return None
        else:
            res_idx = binary_search(
                lambda idx: self.GetRankAtColumnIndex(idx) >= rank,
                0,
                self.GetNumStrataRetained() - 1,
            )
            if res_idx is None:
                return None
            elif self.GetRankAtColumnIndex(res_idx) == rank:
                return res_idx
            else:
                return None

    def DelRanks(
        self: 'HereditaryStratumOrderedStoreList',
        ranks: typing.Iterator[int],
        # deposition ranks might not be stored in strata
        get_column_index_of_rank: typing.Optional[typing.Callable]=None,
    ) -> None:
        """Purge strata with specified deposition ranks from the store.

        Parameters
        ----------
        ranks : iterator over int
            The ranks that to be deleted.
        get_column_index_of_rank : callable, optional
            Callable that returns the deposition rank of the stratum positioned
            at index i among retained strata.
        """

        if get_column_index_of_rank is None:
            get_column_index_of_rank = self.GetColumnIndexOfRank

        indices = [
            get_column_index_of_rank(rank)
            for rank in ranks
        ]
        # adapted from https://stackoverflow.com/a/11303234/17332200
        # iterate over indices in reverse order to prevent invalidation
        # reversed() is an potential optimization
        # given indices is assumed to be in ascending order
        for index in sorted(reversed(indices), reverse=True):
            assert index is not None
            del self._data[index]

    def GetRetainedRanks(
        self: 'HereditaryStratumOrderedStoreDict',
    ) -> typing.Iterator[int]:
        """Get an iterator over deposition ranks of strata present in the
        store.

        Order of iteration should not be considered guaranteed. The store may
        be altered during iteration without iterator invalidation, although
        subsequent updates will not be reflected in the iterator.
        """

        # must make copy to prevent invalidation when strata are deleted
        # note, however, that copy is made lazily
        # (only when first item requested)
        ranks = [
            stratum.GetDepositionRank()
            for stratum in self._data
        ]
        for rank in ranks:
            assert rank is not None
            yield rank

    def IterRankDifferentia(
        self: 'HereditaryStratumOrderedStoreList',
        # deposition ranks might not be stored in strata
        get_rank_at_column_index: typing.Optional[typing.Callable]=None,
        start_column_index: int=0,
    ) -> typing.Iterator[typing.Tuple[int, int]]:
        """Get an iterator over tuples containing deposition ranks and
        differentia of retained strata.

        Guaranteed ordered from most ancient to most recent.

        Parameters
        ----------
        get_rank_at_column_index : callable, optional
            Callable that returns the deposition rank of the stratum positioned
            at index i among retained strata.
        start_column_index : callable, optional
            Number of strata to skip over before yielding first result from the
            iterator. Default 0, meaning no strata are skipped over.
        """

        if get_rank_at_column_index is None:
            get_rank_at_column_index = self.GetRankAtColumnIndex

        # adapted from https://stackoverflow.com/a/12911454
        for index in range(start_column_index, len(self._data)):
            stratum = self._data[index]
            yield (get_rank_at_column_index(index), stratum.GetDifferentia())

    def Clone(
            self: 'HereditaryStratumOrderedStoreList',
    ) -> 'HereditaryStratumOrderedStoreList':
        """Create a copy of the store with identical data that may be freely
        altered without affecting data within this store."""

        # shallow copy
        result = copy(self)
        # do semi-shallow clone on select elements
        # see https://stackoverflow.com/a/47859483 for performance consierations
        result._data = [*self._data]
        return result



import typing

from ..HereditaryStratum import HereditaryStratum
from ..stratum_retention_predicates \
    import StratumRetentionPredicatePerfectResolution


class StratumRetentionCondemnerPerfectResolution(
    # inherit CalcNumStrataRetainedUpperBound, etc.
    StratumRetentionPredicatePerfectResolution,
):
    """Functor to implement the perfect resolution stratum retention policy, for
    use with HereditaryStratigraphicColumn.

    This functor enacts the perfect resolution policy by specifying the set of
    strata ranks that should be purged from a hereditary stratigraphic column
    when the nth stratum is deposited.

    The perfect resolution policy retains all strata. So, comparisons between
    two columns under this policy will detect MRCA rank with zero
    uncertainty. So, MRCA rank estimate uncertainty scales as O(1) with respect
    to the greater number of strata deposited on either column.

    Under the perfect resolution policy, the number of strata retained (i.e.,
    space complexity) scales as O(n) with respect to the number of strata
    deposited.

    See Also
    --------
    StratumRetentionPredicatePerfectResolution:
        For definitions of methods inherited by this class that describe
        guaranteed properties of the perfect resolution stratum retention
        policy.
    """

    def __call__(
        self: 'StratumRetentionCondemnerPerfectResolution',
        retained_ranks: typing.Optional[typing.Iterable[int]]=None,
        num_stratum_depositions_completed: typing.Optional[int]=None,
    ) -> typing.Iterator[int]:
        """Decide which strata within the stratagraphic column should be purged.

        Every time a new stratum is deposited, this method is called to
        determine which strata should be purged. All strata at ranks yielded
        from this functor are immediately purged from the column, meaning that
        for a stratum to persist it must not be yielded by this functor each
        and every time a new stratum is deposited.

        Parameters
        ----------
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.
        retained_ranks : iterator over int, optional
            An iterator over ranks of strata currently retained within the
            hereditary stratigraphic column. Not used in this functor.

        Returns
        -------
        iterator over int
                The ranks of strata that should be purged from the hereditary
            stratigraphic column at this deposition step.

        See Also
        --------
        StratumRetentionPredicatePerfectResolution:
            For details on the rationale, implementation, and guarantees of the
            perfect resolution stratum retention policy.
        """

        # empty generator
        # see https://stackoverflow.com/a/13243870/17332200
        return
        yield



import random
import typing


class StratumRetentionPredicateStochastic:
    """Functor to implement the stochastic resolution stratum retention policy,
    for use with HereditaryStratigraphicColumn.

    This functor enacts the stochastic resolution policy by specifying
    whether a stratum with deposition rank r should be retained within the
    hereditary stratigraphic column after n strata have been deposited.

    The stochastic resolution policy retains strata probabilistically. It would
    be a poor choice to use in practice because mismatches between the
    particular ranks that each column happens to have strata for will degrade
    the effectiveness of comparisons between columns. Rather, it is included in
    the library as an edge case for testing purposes. Worst-case MRCA rank estimate uncertainty scales as O(n) with respect to the greater number of strata deposited on either column being compared.

    Under the stochastic resolution policy, the worst and average case number
    of strata retained (i.e., space complexity) scales as O(n) with respect to
    the number of strata deposited.
    """

    def __call__(
        self: 'StratumRetentionPredicateStochastic',
        stratum_rank: int,
        num_stratum_depositions_completed: int,
    ) -> bool:
        """Decide if a stratum within the stratagraphic column should be
        retained or purged.

        Every time a new stratum is deposited, this method is called on each
        stratum present in a HereditaryStratigraphicColumn to determine whether
        it should be retained. Strata that return False are immediately purged
        from the column, meaning that for a stratum to persist it must earn a
        True result from this method each and every time a new stratum is
        deposited.

        This functor's retention policy implementation guarantees that the most
        ancient and most recent strata will always be retained. For the
        secondmost recently deposited sratum, a pseudorandom coin flip is
        performed. Depending on the outcome of that coin flip, the stratum is either immediately purged or retained permanently.

        Parameters
        ----------
        stratum_rank : int
            The number of strata that were deposited before the stratum under
            consideration for retention.
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.

        Returns
        -------
        bool
            True if the stratum should be retained, False otherwise.
        """

        if (
            stratum_rank
            and stratum_rank == num_stratum_depositions_completed - 2
        ):
            return random.choice([True, False])
        else: return True

    def __eq__(
        self: 'StratumRetentionPredicateStochastic',
        other: 'StratumRetentionPredicateStochastic',
    ) -> bool:
        """Compare for value-wise equality."""

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def CalcNumStrataRetainedUpperBound(
        self: 'StratumRetentionPredicateStochastic',
        num_strata_deposited: int,
    ) -> int:
        """At most, how many strata are retained after n deposted? Inclusive."""

        return num_strata_deposited

    def CalcMrcaUncertaintyUpperBound(
        self: 'StratumRetentionPredicateStochastic',
        *,
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
        actual_rank_of_mrca: typing.Optional[int]=None,
    ) -> int:
        """At most, how much uncertainty to estimate rank of MRCA? Inclusive."""

        return max(
            first_num_strata_deposited,
            second_num_strata_deposited,
        )
